/*
 * Copyright 2006-2008 Kees de Kooter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package editing;

import java.awt.*;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.prefs.Preferences;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;

import javax.swing.text.*;
import javax.swing.text.Segment;
import javax.swing.text.Utilities;



public class SolView extends WrappedPlainView {

    private static HashMap<Pattern, TextAttr> patternColors;
	private static String QUOTE = "(\".*?\")";
	private static String QUOTE2 = "(\'.*?\')";
	private static String COMMENT = "(\\-\\-.*)";
	private static String operators = "[\\%\\`e\\-\\s\\*\\+\\/\\[\\{\\^\\(\\)\\[\\]\\{\\}\\:]";
	private static String number = "([0-9]+\\.?[0-9]*)";
	private static String sep = "[\\s\\-\\*\\+\\/\\[\\{\\^\\]\\}\\(\\)\\:]";
    private static String qops = "if|then|else|elseif|do|end|for|while|break|return";
    public static String url = "\u219D(http\\:\\/\\/[\\d\\w\\S/]*)";
    private static String NUMBERS =  operators + number;
	private static String NUMBERS2 = "^[\\s]*" + number;
    private static String OUTPUTLINE = "(\\=\\>)";
    private static String NORMAL = "(\n[^=\n][^->][^\u219D\n]*)";

	public static boolean syntaxHighlight = true;
	public static TextAttr taString;
	public static TextAttr taNumber;
	public static TextAttr taFunction;
	public static TextAttr taOp;
	public static TextAttr taComment;
    public static TextAttr taOutput;
    public static TextAttr taNormal;
    public static TextAttr taSquare;
    public static TextAttr taURL;
    private static Map hints;
    static {

		Preferences prefs = Preferences.userNodeForPackage(SolView.class);
		taString = TextAttr.fromString(prefs.get("TASTRING", "#640064,0"));
		taNumber = TextAttr.fromString(prefs.get("TANUMBER", "#0000FF,0"));
		taFunction = TextAttr.fromString(prefs.get("TAFUNCTION", "#000000,1"));
		taOp = TextAttr.fromString(prefs.get("TAOP", "#FF0000,1"));
		taComment = TextAttr.fromString(prefs.get("TACOMMENT", "#4B91D7,1"));
        taOutput = TextAttr.fromString("#AAAAAA,1");
        taSquare = TextAttr.fromString("#99CC00,2");
        taOutput = TextAttr.fromString("#000000,1");
        taNormal = TextAttr.fromString("#555555,0");
        taURL = TextAttr.fromString("#0000FF,0");

		syntaxHighlight = prefs.getBoolean("TA", true);
        patternColors = new LinkedHashMap<Pattern, TextAttr>();

        
        patternColors
                .put(Pattern.compile(url), taURL);
		patternColors
                .put(Pattern.compile(QUOTE), taString);
		patternColors
                .put(Pattern.compile(QUOTE2), taString);
		patternColors
                .put(Pattern.compile(NUMBERS), taNumber);
        patternColors
                .put(Pattern.compile(NUMBERS2), taNumber);
        
		patternColors
                .put(Pattern.compile(COMMENT), taComment);
        patternColors
                .put(Pattern.compile(OUTPUTLINE), taOutput);
        patternColors
                .put(Pattern.compile(NORMAL), taNormal);
        
        
		patternColors.put(Pattern.compile("([\\*\\+\\-\\/\\=\\^\\.\\#\\~\\`\\:\\;])"), taOp);
        patternColors.put(Pattern.compile("([\\[\\]\\{\\}])"), TextAttr.fromString("#000000,1"));

        String[] ops = qops.split("\\|");

        for (String op : ops) {
            addKeyword(op.trim());
        }
        
        hints = new HashMap();
        hints.put(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		hints.put(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
    }

	public static void addKeyword(String keyword) {
		patternColors.put(Pattern.compile(sep + "(" + keyword + ")" + sep), taFunction);
		patternColors.put(Pattern.compile("^(" + keyword + ")" + sep), taFunction);
		patternColors.put(Pattern.compile(sep + "(" + keyword + ")$"), taFunction);
	}

	

    public SolView(Element element) {

        super(element);
        // Set tabsize to 4 (instead of the default 8)
        getDocument().putProperty(PlainDocument.tabSizeAttribute, 4);
        getDocument().addDocumentListener(new DocumentListener() {
            public void insertUpdate(DocumentEvent de) {
                startMap = null;
            }

            public void removeUpdate(DocumentEvent de) {
                startMap = null;
            }

            public void changedUpdate(DocumentEvent de) {
                startMap = null;
            }

        });
		
    }

    SortedMap<Integer, Integer> startMap;
    SortedMap<Integer, TextAttr> colorMap;
    Document doc;

    private void parse() throws BadLocationException {
        doc = getDocument();
        String text = doc.getText(0, doc.getLength());
        

        startMap = new TreeMap<Integer, Integer>();
        colorMap = new TreeMap<Integer, TextAttr>();

        for (Map.Entry<Pattern, TextAttr> entry : patternColors.entrySet()) {
            if (entry.getKey().toString().equals(NORMAL) && Boolean.TRUE.equals(
                doc.getProperty("IGNORE_NORMAL")))
                continue;
            
            Matcher matcher = entry.getKey().matcher(text);

            while (matcher.find()) {
				int start = matcher.start(1);
				int end = matcher.end(1);
				boolean ignore = false;
				for (Iterator<Entry<Integer, Integer>> entry2 = startMap.entrySet().iterator(); entry2.hasNext();) {
					Entry<Integer, Integer> old = entry2.next();
					if ((old.getKey() >= start) && (old.getValue() <= end)) {
						entry2.remove();
						colorMap.remove(new Integer(start));
					}
					else if ((old.getKey() <= start) && (old.getValue() >= end))
						ignore = true;
				}

				if (ignore) continue;
                startMap.put(start, end);
                colorMap.put(start, entry.getValue());
            }
        }

    }

    
    
    @Override
    protected int drawUnselectedText(Graphics graphics, int x, int y, int p0,
            int p1) throws BadLocationException {
        
        
        Graphics2D g2 = (Graphics2D) graphics;
        g2.addRenderingHints(hints);
        
        boolean doSyntaxHighlight = syntaxHighlight;
		try {
        
        Segment segment = getLineBuffer();
        if (startMap == null)
            parse();

		int i = p0;

        
		for (Map.Entry<Integer, Integer> entry : startMap.entrySet()) {
            int start = entry.getKey();
            int end = entry.getValue();
			if ((start > p1) || (end < p0))
				continue;
			
			if (end >= p1)
				end = p1;
			if (start <= p0)
				start = p0;

            if (i < start) {                
				graphics.setColor(Color.black);
                doc.getText(i, start - i, segment);

                x = Utilities.drawTabbedText(segment, x, y, graphics, this, i);
            }



            graphics.setColor(colorMap.get(entry.getKey()).color);
			int style = colorMap.get(entry.getKey()).style;
			if (style != Font.PLAIN)
				graphics.setFont(graphics.getFont().deriveFont(style));
            i = end;
            
			doc.getText(start, end-start, segment);

            x = Utilities.drawTabbedText(segment, x, y, graphics, this, start);
			if (style != Font.PLAIN)
				graphics.setFont(graphics.getFont().deriveFont(Font.PLAIN));
        }
		
        // Paint possible remaining text black
        if (i < p1) {
            graphics.setColor(Color.black);
            doc.getText(i, p1 - i, segment);
            x = Utilities.drawTabbedText(segment, x, y, graphics, this, i);
        }

		} catch (Exception e) {
			e.printStackTrace();
		}
        return x;
    }

}
