//
//  Integration.c
//  Systemic Framework
//
//

#include <gsl/gsl_matrix_double.h>

#include "integration.h"
#include "swift.h"

double ok_min_distance = RJUP / AU;
ok_system* ok_alloc_system(int nplanets) {
    ok_system* system = (ok_system*) malloc(sizeof(ok_system));
    system->nplanets = nplanets;
    system->epoch = INVALID_NUMBER;
    system->time = INVALID_NUMBER;
    system->elements = gsl_matrix_calloc(nplanets + 1, ELEMENTS_SIZE);
    MSET(system->elements, 0, MASS, 1.);
    system->xyz = NULL;
    system->orbits = NULL;
    system->flag = 0;
    return system;
}

void ok_free_system(ok_system* system) {
    if (system == NULL)
        return;
    
    gsl_matrix_free(system->elements);
    if (system->xyz != NULL)
        gsl_matrix_free(system->xyz);
    if (system->orbits != NULL)
        gsl_matrix_free(system->orbits);
    
    free(system);
}

ok_system* ok_copy_system(const ok_system* orig) {
    ok_system* system = ok_alloc_system(orig->nplanets);
    gsl_matrix_memcpy(system->elements, orig->elements);
    if (orig->orbits != NULL)
        system->orbits = ok_matrix_copy(orig->orbits);
    if (orig->xyz != NULL)
        system->xyz = ok_matrix_copy(orig->xyz);
    system->epoch = orig->epoch;
    system->time = orig->time;
    return system;
}


void ok_copy_system_to(const ok_system* orig, ok_system* dest) {
    gsl_matrix_memcpy(dest->elements, orig->elements);
    if (orig->orbits != NULL) {
        if (dest->orbits == NULL)
            dest->orbits = ok_matrix_copy(orig->orbits);
        else
            gsl_matrix_memcpy(dest->orbits, orig->orbits);
    }
    if (orig->xyz != NULL) {
        if (dest->xyz == NULL)
            dest->xyz = ok_matrix_copy(orig->xyz);
        else
            gsl_matrix_memcpy(dest->xyz, orig->xyz);
    }
    dest->epoch = orig->epoch;
    dest->time = orig->time;
}


void ok_resize_system(ok_system* system, int npnew) {
    
    gsl_matrix* old = system->elements;
    assert(old != NULL);
    assert(npnew >= 1);
    
    gsl_matrix* new = gsl_matrix_calloc(npnew, ELEMENTS_SIZE); 
    for (int i = 0; i < min(npnew, old->size1); i++) {
        for (int j = 0; j < ELEMENTS_SIZE; j++)
            MSET(new, i, j, MGET(old, i, j));
    }
    
    gsl_matrix_free(old);
    system->elements = new;
}

double ok_acalc(double Mcenter, double Mp, double P) {
    double a = cbrt(P * P * (Mcenter + Mp) / (4.*M_PI*M_PI));
    return a;
}

double ok_pcalc(double Mcenter, double Mp, double a) {
    return sqrt(a * a * a / (Mcenter + Mp)) * 2. * M_PI;
}

void ok_setup(ok_system* system) {
    if (system->orbits != NULL && MROWS(system->orbits) != system->nplanets + 1) {
        gsl_matrix_free(system->orbits);
        system->orbits = NULL;
    }
    if (system->orbits == NULL)
        system->orbits = gsl_matrix_calloc(system->nplanets + 1, ELEMENTS_SIZE);
    if (system->xyz != NULL && MROWS(system->xyz) != system->nplanets + 1) {
        gsl_matrix_free(system->xyz);
        system->xyz = NULL;
    }
    if (system->xyz == NULL)
        system->xyz = gsl_matrix_calloc(system->nplanets + 1, 7);
    
    
    // Copy elements matrix into orbits matrix (for instance, to carry on flags
    // such as ORD)
    gsl_matrix_memcpy(system->orbits, system->elements);
    
    double Mcenter = MSUN_TO_INT(MGET(system->elements, 0, MASS));

    MSET(system->orbits, 0, MASS, Mcenter);
    
    for (int i = 1; i < system->nplanets + 1; i++) {
        double mass = MJUP_TO_INT(MGET(system->elements, i, MASS));
        double per = MGET(system->elements, i, PER);
        double a = ok_acalc(Mcenter, mass, per);
        
        double e = MGET(system->elements, i, ECC);
        double lop = MGET(system->elements, i, LOP);
        double ma = MGET(system->elements, i, MA);
        double node = MGET(system->elements, i, NODE);
        double inc = MGET(system->elements, i, INC);
        
        MSET(system->orbits, i, PER, per);
        MSET(system->orbits, i, MASS, mass);
        MSET(system->orbits, i, MA, TO_RAD(ma));
        MSET(system->orbits, i, ECC, e);
        MSET(system->orbits, i, LOP, TO_RAD(lop));
        MSET(system->orbits, i, INC, TO_RAD(inc));
        MSET(system->orbits, i, NODE, TO_RAD(node));
        MSET(system->orbits, i, SMA, a);
    }
    
    system->time = system->epoch;
    ok_el2cart(system, system->xyz);
        
}



void ok_el2cart(ok_system* system, gsl_matrix* xyz) {

    
    gsl_matrix_set_zero(xyz);
    MSET(xyz, 0, 0, MGET(system->orbits, 0, MASS));
    
    for (int i = 1; i < system->nplanets + 1; i++) {
        double mass = MGET(system->orbits, i, MASS);
        double mu = mass + MGET(xyz, 0, 0);
        double e = MGET(system->orbits, i, ECC);
        double q = MGET(system->orbits, i, SMA) * (1. - e);
        double inc = MGET(system->orbits, i, INC);
        double ma = MGET(system->orbits, i, MA);
        double node = MGET(system->orbits, i, NODE);
        double lop = MGET(system->orbits, i, LOP);    
        double x, y, z, u, v, w;
        
        mco_el2x__(&mu, &q, &e, &inc, &lop, &node, &ma,
                   &x, &y, &z, &u, &v, &w);
        
        MSET(xyz, i, 0, mass);
        MSET(xyz, i, X, x);
        MSET(xyz, i, Y, y);
        MSET(xyz, i, Z, z);
        MSET(xyz, i, VX, u);
        MSET(xyz, i, VY, v);
        MSET(xyz, i, VZ, w);
    }
    

}

void ok_to_cm(ok_system* system, gsl_matrix* xyz) {
    double com[7] = {0.,0.,0.,0.,0.,0.,0.};
    
    for (int i = 0; i < system->nplanets + 1; i++) {
        com[0] += MGET(xyz, i, 0);
        for (int j = 1; j < 7; j++)
            com[j] += MGET(xyz, i, 0) * MGET(xyz, i, j);
    }
    
    for (int i = 0; i < system->nplanets + 1; i++) {
        for (int j = 1; j < 7; j++)
            MINC(xyz, i, j, - com[j]/com[0]);
    }   
}

void ok_to_star(ok_system* system, gsl_matrix* xyz) {
    for (int i = 1; i < system->nplanets + 1; i++)
        for (int j = 1; j < 7; j++)
            MINC(xyz, i, j, -MGET(xyz, 0, j));
    for (int j = 1; j < 7; j++)
        MSET(xyz, 0, j, 0.);
}

void ok_cart2el(ok_system* system, gsl_matrix* els, bool internal) {
    
    gsl_matrix* xyz = system->xyz;
    
    for (int i = 1; i < system->nplanets + 1; i++) {
        double mass = MGET(xyz, i, 0);
        double mu = mass + MGET(xyz, 0, 0);
        double x = MGET(xyz, i, X) - MGET(xyz, 0, X);
        double y = MGET(xyz, i, Y) - MGET(xyz, 0, Y);
        double z = MGET(xyz, i, Z) - MGET(xyz, 0, Z);
        double u = MGET(xyz, i, VX) - MGET(xyz, 0, VX);
        double v = MGET(xyz, i, VY) - MGET(xyz, 0, VY);
        double w = MGET(xyz, i, VZ) - MGET(xyz, 0, VZ);
        double q, e, inc, p, n, l;
        
        mco_x2el__(&mu, &x, &y, &z, &u, &v, &w, &q, &e, &inc, &p, &n, &l);
        
        double a = q / (1-e);
        double P = ok_pcalc(MGET(xyz, 0, 0), mass, a);
        
        MSET(els, i, PER, P);
        MSET(els, i, MASS, mass);
        MSET(els, i, MA, l);
        MSET(els, i, ECC, (e > 1e-6 ? e : 0.));
        MSET(els, i, INC, inc);
        MSET(els, i, NODE, n);
        MSET(els, i, SMA, a);
        MSET(els, i, LOP, p);
    }
    
}

int ok_force(double t, const double y[], double f[], void* params) {
    ok_system* system = (ok_system*) params;
    
    const int N = system->nplanets + 1;
    
    for (int i = 0; i < 7 * N; i++)
        f[i] = 0.;
    
    for (int i = 0; i < N; i++) {
        f[i * 7 + 1] = y[i * 7 + 4];
        f[i * 7 + 2] = y[i * 7 + 5];
        f[i * 7 + 3] = y[i * 7 + 6];
        
        for (int j = i + 1; j < N; j++) {
            double m1 = y[i * 7];
            double m2 = y[j * 7];
            
            double i_rsq = 1./(sqr(y[i * 7 + 1] - y[j * 7 + 1]) +
                sqr(y[i * 7 + 2] - y[j * 7 + 2]) +
                sqr(y[i * 7 + 3] - y[j * 7 + 3]));
            
            double i_r = sqrt(i_rsq);
            
            if (i_r < ok_min_distance) {
                return GSL_ERANGE;
            }
            
            double a1 = -m2 * i_rsq * i_r;
            double a2 = -m1 * i_rsq * i_r;
            
            for (int d = 1; d <= 3; d++) {
                double w = (y[i * 7 + d] - y[j * 7 + d]);
                f[i * 7 + d + 3] += w * a1;
                f[j * 7 + d + 3] -= w * a2;
            }
        }        
    }


    return GSL_SUCCESS;
}


int ok_jac(double t, const double y[], double *dfdy, 
           double dfdt[], void *params) {
    ok_system* system = (ok_system*) params;
    const int N = system->nplanets + 1;
    
    gsl_matrix_view dfdy_mat = gsl_matrix_view_array (dfdy, 7 * N, 7 * N);
    gsl_matrix * jac = &dfdy_mat.matrix; 
    gsl_matrix_set_zero(jac);
    
    for (int i = 0; i < N; i++) {
        // the N * 7-th row [mass row] is always 0
        
        // df_x / dx = 0
        // df_x_i / dv_i = 1        
        MSET(jac, i * 7 + 1, i * 7 + 4, 1);
        MSET(jac, i * 7 + 2, i * 7 + 5, 1);
        MSET(jac, i * 7 + 3, i * 7 + 6, 1);        
        
        // df_v / dv = 0
        // df_v_i / dx =
        
        for (int j = i+1; j < N; j++) {
            const double inv_rij = 1./sqrt(sqr(y[i * 7 + 1] - y[j * 7 + 1]) + 
                                        sqr(y[i * 7 + 2] - y[j * 7 + 2]) + 
                                        sqr(y[i * 7 + 3] - y[j * 7 + 3]));
            const double inv_rij_3 = inv_rij * inv_rij * inv_rij;
            const double inv_rij_5 = inv_rij_3 * inv_rij * inv_rij;
            
            for (int m = 1; m <= 3; m++) 
                for (int n = 1; n <= 3; n++) {
                    double term = - (y[i * 7 + m] - y[j * 7 + m])
                         * (y[i * 7 + n] - y[j * 7 + n]) * inv_rij_5;
                    
                    if (m == n)
                        term += - inv_rij_3;
                    
                    MINC(jac, i * 7 + m + 3, i * 7 + n, y[j * 7] * term);
                    MINC(jac, j * 7 + m + 3, j * 7 + n, - y[i * 7] * term);   
                }

        }
    }
    return GSL_SUCCESS;
}

static ok_integrator_options defoptions = { 1e-11, 1e-11, &ok_force, &ok_jac };

/// This routine integrates the system in time. An array of snapshots taken at each time specified
/// by the times vector is returned. 
ok_system** ok_integrate_gsl(ok_system* initial, const gsl_vector* times, const ok_integrator_options* options, const gsl_odeiv_step_type * solver,
        ok_system** bag) {
    
    // Check that the system has been set-up
    assert(initial->xyz != NULL);
    // Check input arguments
    assert(times != NULL);
    assert(solver != NULL);
    
    const double startTime = initial->epoch;
    const int NDIMS = initial->nplanets + 1;
    
    // Allocate the return array of snapshots
    const int SAMPLES = times->size;
   
    if (bag == NULL) {
        bag = (ok_system**) calloc(SAMPLES, sizeof(ok_system*));
        for (int i = 0; i < SAMPLES; i++) {
            bag[i] = ok_copy_system(initial);
            bag[i]->epoch = initial->epoch;
            bag[i]->xyz = (bag[i]->xyz != NULL ? bag[i]->xyz : gsl_matrix_alloc(initial->nplanets+1, 7));
        }
    } else {
        for (int i = 0; i < SAMPLES; i++) {
            bag[i]->epoch = initial->epoch;
            bag[i]->flag = initial->flag;
            gsl_matrix_memcpy(bag[i]->elements, initial->elements);
            gsl_matrix_memcpy(bag[i]->orbits, initial->orbits);
            bag[i]->xyz = (bag[i]->xyz != NULL ? bag[i]->xyz : gsl_matrix_alloc(initial->nplanets+1, 7));
        }
    }

    
    // Initialize the GSL structures to solve the ODE.
    gsl_odeiv_step *stepper = gsl_odeiv_step_alloc(solver, NDIMS * 7);
    gsl_odeiv_control * control = gsl_odeiv_control_y_new(options->abs_acc, options->rel_acc);
    gsl_odeiv_evolve * e = gsl_odeiv_evolve_alloc (NDIMS * 7);
    
    gsl_odeiv_system eqns;
    eqns.dimension = NDIMS * 7;
    eqns.function = options->force;
    eqns.jacobian = options->jac;
    eqns.params = initial;
    
    // Allocates the temporary buffer to hold cartesian coordinates
    double prevTime = startTime;
    gsl_matrix* prevOrbits = initial->orbits;
    
    double h = 0.01;
    gsl_matrix* xyz = gsl_matrix_alloc(initial->nplanets+1, 7);
    gsl_matrix_memcpy(xyz, initial->xyz);
    
    // Loop through the times vector
    for (int i = 0; i < SAMPLES; i++) {
        double time = times->data[i];
        

        // Integrate between prevTime and time
        if (fabs(time - prevTime) > 1e-10) {
            while (fabs(time - prevTime) > 1e-10) {
                h = copysign(h, time-prevTime);
                const int result = gsl_odeiv_evolve_apply(e, control, stepper, &eqns, &prevTime, time, &h, xyz->data);

                if (result != GSL_SUCCESS) {
                    // TODO: Fail gracefully
                    gsl_matrix_free(xyz);
                    for (int i = 0; i < SAMPLES; i++)
                        ok_free_system(bag[i]);
                    free(bag);
                    fprintf(stderr, "Error integrating.\n");
                    break;
                }
            }
        } else {
            if (i == 0)
                gsl_matrix_memcpy(xyz, initial->xyz);
            else
                gsl_matrix_memcpy(xyz, bag[i-1]->xyz);
        };
        
        
        // Set up the return vector
        bag[i]->time = bag[i]->epoch = time;
       
        gsl_matrix_memcpy(bag[i]->xyz, xyz);
        gsl_matrix_memcpy(bag[i]->orbits, prevOrbits);
        
        ok_cart2el(bag[i], bag[i]->orbits, true);
        prevOrbits = bag[i]->orbits;
        
        
        // Ensures that the force/jacobian routines are passed the most recent state of
        // the system
        eqns.params = bag[i];
        
        prevTime = time;
    }
    
    gsl_odeiv_control_free(control);
    gsl_odeiv_evolve_free(e);
    gsl_odeiv_step_free(stepper);
    
    gsl_matrix_free(xyz);
    return bag;
}

ok_system** ok_integrate_kep(ok_system* initial, const gsl_vector* times, 
                             const ok_integrator_options* options, ok_system** bag) {
    
    assert(initial->orbits != NULL);
    
    const double startTime = initial->epoch;
    const int NDIMS = initial->nplanets + 1;
    
    // Allocate the return array of snapshots
    const int SAMPLES = times->size;
    if (bag == NULL) {
        bag = (ok_system**) calloc(SAMPLES, sizeof(ok_system*));
        for (int i = 0; i < SAMPLES; i++) {
            bag[i] = ok_copy_system(initial);
            bag[i]->epoch = initial->epoch;
            bag[i]->xyz = (bag[i]->xyz != NULL ? bag[i]->xyz : gsl_matrix_alloc(initial->nplanets+1, 7));
        }
    } else {
        for (int i = 0; i < SAMPLES; i++) {
            bag[i]->epoch = initial->epoch;
            bag[i]->flag = initial->flag;
            gsl_matrix_memcpy(bag[i]->elements, initial->elements);
            gsl_matrix_memcpy(bag[i]->orbits, initial->orbits);
            bag[i]->xyz = (bag[i]->xyz != NULL ? bag[i]->xyz : gsl_matrix_alloc(initial->nplanets+1, 7));
        }
    }
    
    double prevTime = startTime;
    ok_system* prevSystem = initial;
    
    double n[NDIMS];
    for (int i = 1; i < NDIMS; i++)
        n[i] = 2*M_PI/MGET(prevSystem->orbits, i, PER);
    
    // Loop through the times vector
    for (int i = 0; i < SAMPLES; i++) {
        
        double dt = times->data[i] - prevTime;
        for (int j = 1; j < NDIMS; j++) {
            double ma = RADRANGE(MGET(prevSystem->orbits, j, MA) + n[j] * dt);
            MSET(bag[i]->orbits, j, MA, ma);
        }
        
        bag[i]->time = times->data[i];
        bag[i]->epoch = bag[i]->time;
        ok_el2cart(bag[i], bag[i]->xyz);
        prevTime = times->data[i];
        prevSystem = bag[i];
    }
    
    return bag;
}

ok_system** ok_integrate(ok_system* initial, const gsl_vector* times, const ok_integrator_options* options, const int integrator,
        ok_system** bag) {
    // Check that the system has been set-up
    assert(initial->xyz != NULL);
    // Check input arguments
    assert(times != NULL);
    
    assert(! IS_INVALID(initial->epoch));
    if (IS_INVALID(initial->time))
        initial->time = initial->epoch;
    
    if (options == NULL)
        options = &defoptions;
    
    switch (integrator) {
        case KEPLER:
            return ok_integrate_kep(initial, times, options, bag);
            break;
        case RK5:
            return ok_integrate_gsl(initial, times, options, gsl_odeiv_step_rkf45, bag);
            break;
        case RK8:
            return ok_integrate_gsl(initial, times, options, gsl_odeiv_step_rk8pd, bag);
            break;
        case SWIFT_BS:
            return ok_integrate_swift(initial, times, options, 0, bag);
        case SWIFT_WHM:
            return ok_integrate_swift(initial, times, options, 1, bag);
        case SWIFT_RMVS3:
            return ok_integrate_swift(initial, times, options, 2, bag);
    }
    
    
    return NULL;
}



double ok_get_rv(ok_system* sys) {
    assert(sys->xyz != NULL);    
    ok_to_cm(sys, sys->xyz);
    double rv = -MGET(sys->xyz, 0, VZ);
    rv = AUPDAY_TO_MPS(rv);
    return rv;
}

gsl_matrix* ok_get_rvs(ok_system** sys, int len) {
    gsl_matrix* ret = gsl_matrix_alloc(len, 2);
    
    for (int i = 0; i < len; i++) {
        MSET(ret, i, 0, sys[i]->time);
        MSET(ret, i, 1, ok_get_rv(sys[i]));
    }
    return ret;
}

gsl_matrix* ok_get_xyzs(ok_system** bag, int len) {
    int N = bag[0]->nplanets + 1;
    gsl_matrix* xyzs = gsl_matrix_calloc(len, 7 * N + 1);
    
    for (int i = 0; i < len; i++) {
        assert(bag[i]->xyz != NULL);
        
        MSET(xyzs, i, 0, bag[i]->time);
        
        for (int j = 0; j < N; j++)
            for (int p = 0; p < 7; p++)
                MSET(xyzs, i, j * 7 + p + 1, MGET(bag[i]->xyz, j, p));
    }
    
    return xyzs;
}

gsl_matrix* ok_get_els(ok_system** bag, int len, bool internal) {
    int N = bag[0]->nplanets + 1;
    gsl_matrix* els = gsl_matrix_calloc(len, N * ELEMENTS_SIZE + 1);
    
    for (int i = 0; i < len; i++) {
        assert(bag[i]->orbits != NULL);
        
        MSET(els, i, 0, bag[i]->time);
        
        for (int j = 0; j < N; j++)
            for (int p = 0; p < ELEMENTS_SIZE; p++) {
                double v = MGET(bag[i]->orbits, j, p);
                
                if (p == MASS && !internal)
                    v = INT_TO_MJUP(v);
                else if ((p == MA || p == LOP || p == INC || p == NODE) && !internal)
                    v = TO_DEG(v);
                
                MSET(els, i, j * ELEMENTS_SIZE + p + 1, v);
            }
    }
    
    return els;
}

/*
double ok_find_closest_transit_old(ok_system* sys, const int plidx, const ok_integrator_options* options, const int intMethod, double eps) {
    int pidx = plidx;
    
    for (int i = 1; i < sys->elements->size1; i++) {
        if (pidx == (int) MGET(sys->elements, i, ORD)) {
            pidx = i;
            break;
        }
    }
    
    double period = MGET(sys->elements, pidx, PER);
    if (sys->xyz == NULL) {
        sys->xyz = gsl_matrix_alloc(sys->nplanets + 1, 7);
        ok_el2cart(sys, sys->xyz);
    }
    
    ok_system* s1 = ok_copy_system(sys);
    s1->time = s1->epoch;
    ok_system* s2;
    double t2;
    
    if (MGET(sys->xyz, pidx, X) > 0 && MGET(sys->xyz, pidx, Y) > 0)
        t2 = s1->time + 0.5 * period;
    else if (MGET(sys->xyz, pidx, X) < 0 && MGET(sys->xyz, pidx, Y) > 0)
        t2 = s1->time - 0.5 * period;
    else if (MGET(sys->xyz, pidx, X) < 0 && MGET(sys->xyz, pidx, Y) < 0)
        t2 = s1->time - period;
    else if (MGET(sys->xyz, pidx, X) > 0 && MGET(sys->xyz, pidx, Y) < 0)
        t2 = s1->time + period;
    
    
    ok_system** buf = NULL;
        
    gsl_vector* times = gsl_vector_alloc(3);
    VSET(times, 0, s1->time);
    VSET(times, 1, t2);
    VSET(times, 2, t2);
    
    buf = ok_integrate(s1, times, options, intMethod, buf);
    s2 = ok_copy_system(buf[1]);
    
    ok_to_star(s1, s1->xyz);
    ok_to_star(s2, s2->xyz);
    
    
    if (MGET(s1->xyz, pidx, X) > 0.) {
        ok_system* tmp = s1;
        s1 = s2; s2 = tmp;        
    }
    
    
    
    int nsteps = 0;
    while (fabs(s1->time - s2->time) > eps) {
        double ta = 0.5 * (s1->time + s2->time);
        VSET(times, 0, s1->time);
        VSET(times, 1, ta);
        VSET(times, 2, s2->time);
        
        buf = ok_integrate(s1, times, options, intMethod, buf);
        
        ok_system* candidate = ok_copy_system(buf[1]);
        ok_to_star(candidate, candidate->xyz);
        
        
        
        if (MGET(candidate->xyz, pidx, X) < 0.) {
            ok_free_system(s1);
            s1 = candidate;
            s1->epoch = s1->time;
        } else {
            ok_free_system(s2);
            s2 = candidate;
            s2->epoch = s2->time;
        }
        
        
        nsteps ++;
    }
    
    return 0.5 * (s1->time + s2->time);
} 
*/


/**
 * Returns the time of the transit closest to the time of the current state of the system for the specified planet.
 * @param state pointer to a ok_system structure containing cartesian coords + elements, time and epoch.
 * @param plidx index of the planet (1..n)
 * @param options NULL for the default integrator options, or pass a pointer to an ok_integrator_options structure
 * @param intMethod integration method (one of the builtin integrators: KEPLER, RK5, RK8, SWIFT_BS, SWIFT_RMVS or
 * a custom integrator)
 * @param eps tolerance on the time
 * @return Closest transit to the time specified in state
 */
int ok_find_closest_transit(ok_system* state, const int plidx, const ok_integrator_options* options, const int intMethod, double eps,
        double* timeout) {
    // Find planet tagged with the specified index (ORD column), since internally we keep planets sorted
    // by period.
    int pidx = plidx;
    int retval = OK_SUCCESS;
    
    ok_to_star(state, state->xyz);
    
    for (int i = 1; i < state->elements->size1; i++) {
        if (pidx == (int) MGET(state->elements, i, ORD)) {
            pidx = i;
            break;
        }
    }
    
    double t;
    //if (intMethod == KEPLER) {

        // Calculate approximate time of closest transit. Angles in the orbits matrix are in radians,
        // with epoch stored in state->epoch.
        double aop = MGET(state->orbits, pidx, LOP) - MGET(state->orbits, pidx, NODE);
        double per = MGET(state->orbits, pidx, PER);
        double n = 2*M_PI/per;
        double tau = (n * state->time - MGET(state->orbits, pidx, MA))/n;
        double e = MGET(state->orbits, pidx, ECC);

        // Elliptic expansion; fail if e > 0.6. Should handle this case later.
        if (e > 0.6) {
            *timeout = INVALID_NUMBER;
            return OK_NOCONV;
        }

        double ftra = RADRANGE(0.5 * M_PI - aop);
        double M_1 = RADRANGE(ftra-2.*e*sin(ftra)+0.75*e*e*sin(2*ftra));
        t = M_1/n + tau;

        if (fabs(t - state->time - per) < fabs(t-state->time))
            t -= per;
        else if (fabs(t - state->time + per) < fabs(t-state->time))
            t += per;
        
    //} else {
        
        
        
    //};
    
    
    
    
    //printf("LOP = %e, M = %e, P = %e, e = %e, MASS = %e, X = %e\n", MGET(state->orbits, pidx, LOP), MGET(state->orbits, pidx, MA),
    //      MGET(state->orbits, pidx, PER), MGET(state->orbits, pidx, ECC), MGET(state->orbits, pidx, MASS),
    //        MGET(state->xyz, pidx, X));
    
    // Allocate working space
    const int NDIMS = state->nplanets + 1;
    gsl_matrix* fout = gsl_matrix_alloc(NDIMS, 7);
    gsl_vector* times = gsl_vector_alloc(1);
    
    ok_system** bag = NULL;
    VSET(times, 0, t);
    bag = ok_integrate(state, times, options, intMethod, bag);
    ok_system* buf = ok_copy_system(bag[0]);
    
    double diff = DBL_MAX;
    int steps = 0;
    
    while (diff > eps) {
        // Center on star
        ok_to_star(buf, buf->xyz);
        // Calculate cartesian accelerations
        ok_force(buf->time, buf->xyz->data, fout->data, buf);
        // Calculate r_s derivatives
        //double rs = sqrt(SQR(MGET(buf->xyz, pidx, X)) + SQR(MGET(buf->xyz, pidx, Y)));
        
        double rs_1 = MGET(buf->xyz, pidx, X) * MGET(buf->xyz, pidx, VX) +
                MGET(buf->xyz, pidx, Y) * MGET(buf->xyz, pidx, VY);
        double rs_2 = MGET(buf->xyz, pidx, VX) * MGET(buf->xyz, pidx, VX) +
                MGET(buf->xyz, pidx, X) * MGET(fout, pidx, VX) +
                MGET(buf->xyz, pidx, VY) * MGET(buf->xyz, pidx, VY) +
                MGET(buf->xyz, pidx, Y) * MGET(fout, pidx, VY);
        
        double t_1 = t - rs_1/rs_2;
        diff = fabs(t-t_1);
        
        //printf("%.10e %e %e %e %d %e %e %e\n", t, t-t_1, t - state->time, rs, steps, TO_DEG(MGET(buf->orbits, pidx, MA)),
        //        MGET(buf->xyz, pidx, X), MGET(buf->xyz, pidx, Z));
        
        t = t_1;
        VSET(times, 0, t);
        bag = ok_integrate(buf, times, options, intMethod, bag);
        
        // Set the new system t
        ok_copy_system_to(bag[0], buf);
        
        
        if (steps > 30) {
            
            printf("breaking, %d %e %e %e %e\n", steps, t, sqrt(SQR(MGET(buf->xyz, pidx, X)) + SQR(MGET(buf->xyz, pidx, Y))), diff,
                    MGET(buf->orbits, pidx, SMA));
            retval = OK_NOCONV;
            t = INVALID_NUMBER;
            break;
        }
        
        steps++;
    };
    
    ok_free_system(buf);
    ok_free_system(bag[0]);
    ok_free_system(bag[1]);
    free(bag);
    gsl_matrix_free(fout);
    gsl_vector_free(times);
    
    *timeout = t;
    return retval;
}