/*
 *  kernel.c
 *  Systemic2
 *
 *
 */

#include <gsl/gsl_vector_double.h>
#include <gsl/gsl_matrix_double.h>
#include "mercury.h"
#include "kernel.h"
#include "integration.h"
#include "utils.h"
#include "gsl/gsl_matrix.h"
#include "gsl/gsl_multimin.h"
#include "gsl/gsl_rng.h"
#include "simplex.h"
#include "time.h"

ok_minimizer ok_minimizers[] = { K_minimize_simplex, NULL, NULL, NULL, NULL, NULL };
char * ok_orb_labels[ELEMENTS_SIZE] =  { "P", "M", "MA", "E", "LOP", "I", "NODE", "RADIUS", "A", "K",
        "", "", "", "ORD"};

double ok_default_steps[ELEMENTS_SIZE] = {[PER] = 1e-3, [MASS] = 1e-3, [MA] = 1e-2,
    [ECC] = 1e-2, [LOP] = 1e-2, [INC] = 1e-2, [NODE] = 1e-2, [RADIUS] = 1e-2, [SMA] = -1,
    [SEMIAMP] = -1, 0, 0, 0, 0, 0};

void K_validate(ok_kernel* k);

ok_kernel* K_alloc() {
    ok_kernel* k = (ok_kernel*) malloc(sizeof (ok_kernel));
    memset(k, 0, sizeof(ok_kernel));
    k->chi2 = INVALID_NUMBER;
    k->chi2_rvs = INVALID_NUMBER;
    k->chi2_tts = INVALID_NUMBER;
    k->rms_tts = INVALID_NUMBER;
    k->jitter = INVALID_NUMBER;
    k->rms = INVALID_NUMBER;
    k->nsets = 0;
    k->ndata = 0;
    k->nrvs = 0;
    k->ntts = 0;
    k->intMethod = KEPLER;
    k->Mstar = 1.;
    k->minfunc = K_getChi2;
    k->system = ok_alloc_system(0);
    MSET(k->system->elements, 0, MASS, 1.);
    k->compiled = NULL;
    k->plRanges = (gsl_matrix**) calloc(2, sizeof(gsl_matrix*));
    k->parRanges = (gsl_vector**) calloc(2, sizeof(gsl_vector*));
    
    k->times = NULL;
    k->datasets = (gsl_matrix**) malloc(PARAMS_SIZE * sizeof (gsl_matrix*));
    k->params = gsl_vector_calloc(PARAMS_SIZE);
    k->params->data[NOISE] = 1.;
    
    k->integration = NULL;
    k->plFlags = NULL;
    k->parFlags = gsl_vector_int_calloc(PARAMS_SIZE);
    k->parSteps = gsl_vector_calloc(PARAMS_SIZE);
    VSET(k->parSteps, NOISE, 1e-2);
    
    k->intOptions = NULL;
    k->flags = 0;
    k->ndata = -1;
    k->progress = NULL;
    
    k->rng = gsl_rng_alloc(gsl_rng_default);
    gsl_rng_set(k->rng, clock());
    return k;
}

void K_free(ok_kernel* k) {
    if (k == NULL)
        return;
    
    if (! (k->flags & SHARE_FLAGS)) {
        gsl_matrix_int_free(k->plFlags);
        gsl_vector_int_free(k->parFlags);
    }
    if (! (k->flags & SHARE_STEPS)) {
        gsl_matrix_free(k->plSteps);
        gsl_vector_free(k->parSteps);
        gsl_matrix_free(k->plRanges[0]);
        gsl_matrix_free(k->plRanges[1]);
        gsl_vector_free(k->parRanges[0]);
        gsl_vector_free(k->parRanges[1]);
        free(k->parRanges);
        free(k->plRanges);
    }
    
    
    if (! (k->flags & SHARE_DATA)) {
        for (int i = 0; i < k->nsets; i++) {
            gsl_matrix_free(k->datasets[i]);
        }
        free(k->datasets);
        
        gsl_vector_free(k->times);
        
        if (k->integration != NULL) {
            for (int i = 0; i < k->integrationSamples; i++)
                ok_free_system(k->integration[i]);
            free(k->integration);
        }
        free(k->compiled);
    }
    
    assert(! (k->system->flag & FREED));
    k->system->flag = FREED;
    ok_free_system(k->system);
    gsl_vector_free(k->params);
    
    gsl_rng_free(k->rng);
    k->flags = FREED;
    free(k);
    
}

int K_compile_sort(const void* elem1, const void* elem2) {
    double* e1 = *(double**) elem1;
    double* e2 = *(double**) elem2;
    
    return (e1[T_TIME] - e2[T_TIME] < 0 ? -1 : 1);
}

double** K_compileData(ok_kernel* k) {
    if (k->nsets < 1) {
        k->ndata = 0;
        return NULL;
    }
    if (k->times != NULL)
        gsl_vector_free(k->times);
    
    int ndata = 0;
    int ncols = MCOLS(k->datasets[0]);

    k->nrvs = 0;
    k->ntts = 0;
    
    for (int i = 0; i < k->nsets; i++) {
        ndata += MROWS(k->datasets[i]);
        ncols = MIN(ncols, MCOLS(k->datasets[i]));
        
        if (MROWS(k->datasets[i]) > 0) {
            if ((int) MGET(k->datasets[i], 0, T_FLAG) == T_RV)
                k->nrvs+= MROWS(k->datasets[i]);
            else if ((int) MGET(k->datasets[i], 0, T_FLAG) == T_TIMING)
                k->ntts+= MROWS(k->datasets[i]);
        }
    }
    
    ncols++;

    double** compiled = k->compiled;
    if (k->compiled == NULL || k->ndata != ndata) {
        compiled = (double**) malloc(sizeof(double*) * ndata);
        free(k->compiled);
    }
    
    int nr = 0;

    for (int i = 0; i < k->nsets; i++) {
        int rows = MROWS(k->datasets[i]);
        for (int j = 0; j < rows; j++) {
            compiled[nr] = gsl_matrix_ptr(k->datasets[i], j, 0);
            compiled[nr][T_SET] = (double) i;
            nr++;
        }
    }
    
    if (k->flags & BOOTSTRAP_DATA) {
        k->compiled = (double**) malloc(sizeof(double*) * ndata);
        for (int i = 0; i < ndata; i++) {
            k->compiled[i] = compiled[gsl_rng_uniform_int(k->rng, ndata)];
        }
        free(compiled);
        k->flags &= ~BOOTSTRAP_DATA;
    } else
        k->compiled = compiled;
    
    qsort(compiled, ndata, sizeof(double*), K_compile_sort);
    
    k->times = gsl_vector_calloc(ndata);
    for (int i = 0; i < ndata; i++) {
        VSET(k->times, i, k->compiled[i][T_TIME]);
    }
    
    if (IS_INVALID(k->system->epoch)) {
        k->system->epoch = k->system->time = VGET(k->times, 0);
    }
    k->ndata = ndata;
    
    k->flags &= ~NEEDS_COMPILE;
    return k->compiled;
}

gsl_matrix* K_addDataTable(ok_kernel* k, gsl_matrix* table, const char* name, int type) {
    if (MCOLS(table) < DATA_SIZE) {
        table = ok_matrix_resize(table, MROWS(table), DATA_SIZE);
    }
    
    k->datasets[k->nsets] = table;
    for (int i = 0; i < MROWS(table); i++)
        MSET(table, i, T_FLAG, type);
    
    k->nsets++;
    k->flags |= NEEDS_COMPILE;
    if (type == T_RV)
        VISET(k->parFlags, k->nsets - 1, ACTIVE | MINIMIZE);
    else
        VISET(k->parFlags, k->nsets - 1, ACTIVE);
    return table;
}

gsl_matrix* K_addDataFile(ok_kernel* k, const char* path, int type) {
    FILE* fid = fopen(path, "r");
    assert(fid != NULL);
    if (fid == NULL)
        return NULL;

    k->datasets[k->nsets] = ok_read_table(fid);
    
    for (int i = 0; i < MROWS(k->datasets[k->nsets]); i++)
        MSET(k->datasets[k->nsets], i, T_FLAG, type);
    
    if (type == T_TIMING) {
        /* TODO Accept flag */
        for (int i = 0; i < MROWS(k->datasets[k->nsets]); i++) {
            MSET(k->datasets[k->nsets], i, T_ERR, MGET(k->datasets[k->nsets], i,
                    T_VAL));
            MSET(k->datasets[k->nsets], i, T_VAL, MGET(k->datasets[k->nsets], i,
                    T_TIME)); 
        }
    }
    
    k->nsets++;
    fclose(fid);

    k->flags |= NEEDS_COMPILE;
    if (type == T_RV)
        VISET(k->parFlags, k->nsets - 1, ACTIVE | MINIMIZE);
    else
        VISET(k->parFlags, k->nsets - 1, ACTIVE);
    
    VSET(k->parSteps, k->nsets - 1, 0.05);
    
    return k->datasets[k->nsets-1];
}

void K_removeData(ok_kernel* k, int idx) {
    
    if (idx == -1) {
        for (int i = 0; i < k->nsets; i++) {
            gsl_matrix_free(k->datasets[i]);
            VSET(k->params, i, 0.);
        }
        k->nsets = 0;
        k->ndata = 0;
        k->nrvs = 0;
        k->ntts = 0;
        k->flags |= NEEDS_SETUP;
        return;
    }
    
    gsl_matrix_free(k->datasets[idx]);
    k->datasets[idx] = NULL;
    
    for (int i = idx+1; i < PARAMS_SIZE; i++) {
        k->datasets[i-1] = k->datasets[i];
        k->params->data[i-1] = k->params->data[i];
    }
    
    k->nsets--;
    k->flags |= NEEDS_SETUP;
}

gsl_matrix* K_getData(ok_kernel* k, int idx) {
    return k->datasets[idx];
}


void K_addPlanet(ok_kernel* k, const double elements[]) {

    k->system->nplanets++;
    int row = k->system->nplanets;

    ok_resize_system(k->system, k->system->nplanets + 1);

    k->plFlags = ok_matrix_int_resize(k->plFlags, k->system->nplanets+1, MCOLS(k->system->elements));
    k->plSteps = ok_matrix_resize(k->plSteps, k->system->nplanets+1, MCOLS(k->system->elements));
    
    MSET(k->system->elements, row, ORD, -1);
    MSET(k->system->elements, row, INC, 90.);
    int i = 0;
    while (true) {
        if (elements[i] == DONE)
            break;
        else
            MSET(k->system->elements, row, (int) elements[i], elements[i+1]);
        i+=2;
    }


    if (MGET(k->system->elements, row, ORD) < 0)
        MSET(k->system->elements, row, ORD, row);
    
    
    for (int i = 0; i < 5; i++) 
        MISET(k->plFlags, row, i, ACTIVE | MINIMIZE);
    for (int i = 0; i < ELEMENTS_SIZE; i++) {
        MSET(k->plSteps, row, i, ok_default_steps[i]);
    }
    
    ok_sort_matrix(k->system->elements, ORD);
    k->flags |= NEEDS_SETUP;
    K_validate(k);
}

void K_removePlanet(ok_kernel* k, int idx) {
    
    if (k->system->nplanets == 0)
        return;
    
    gsl_matrix_free(k->system->xyz);
    k->system->xyz = NULL;
    gsl_matrix_free(k->system->orbits);
    k->system->orbits = NULL;
    
    
    gsl_matrix* els = ok_matrix_remove_row(k->system->elements, idx);
    gsl_matrix_free(k->system->elements);
    k->system->elements = els;
    
    gsl_matrix_int* plFlags = ok_matrix_int_remove_row(k->plFlags, idx);
    gsl_matrix_int_free(k->plFlags);
    k->plFlags = plFlags;
    
    if (k->plRanges[0] != NULL) {
        gsl_matrix* plRanges0 = ok_matrix_remove_row(k->plRanges[0], idx);
        gsl_matrix_free(k->plRanges[0]);
        k->plRanges[0] = plRanges0;
        gsl_matrix* plRanges1 = ok_matrix_remove_row(k->plRanges[1], idx);
        gsl_matrix_free(k->plRanges[1]);
        k->plRanges[1] = plRanges1;
    }
    
    for (int i = 0; i < k->system->elements->size1; i++)
        MSET(k->system->elements, i, ORD, i);
    
    k->system->nplanets--;
    k->flags |= NEEDS_SETUP;
}

void K_setElement(ok_kernel* k, int row, int col, double value) {
    
    if (row == -1) {
        for (int i = 1; i < MROWS(k->system->elements); i++)
            K_setElement(k, i, col, value);
    } else {
        if (col >= 0 && col < ELEMENTS_SIZE)
            MSET(k->system->elements, row, col, value);
        else if (col == ECCANOMALY) {
            double e = MGET(k->system->elements, row, ECC);
            double E = TO_RAD(value);
            double M = E-e*sin(E);
            MSET(k->system->elements, row, MA, TO_DEG(M));
        } else if (col == TRUEANOMALY) {
            double e = MGET(k->system->elements, row, ECC);
            double f = RADRANGE(TO_RAD(value));
            double E = 2.*atan(sqrt((1-e)/(1+e)) * tan(0.5*f));
            if (f > 0.5*M_PI && f < 1.5 * M_PI)
                E += M_PI;
            double M = E-e*sin(E);
            MSET(k->system->elements, row, MA, TO_DEG(M));
        }
    }
        
    k->flags |= NEEDS_SETUP;
    K_validate(k);
}

double K_getElement(ok_kernel* k, int row, int col) {
    K_validate(k);
    if (col >= 0 && col < ELEMENTS_SIZE)
        return MGET(k->system->elements, row, col);
    else if (col == ECCANOMALY) {
        double e = MGET(k->system->elements, row, ECC);
        double ma = TO_RAD(MGET(k->system->elements, row, MA));
        double E = mco_kep__(&e, &ma);
        return TO_DEG(E);
    } else if (col == TRUEANOMALY) {
        double e = MGET(k->system->elements, row, ECC);
        double ma = TO_RAD(MGET(k->system->elements, row, MA));
        double E = mco_kep__(&e, &ma);
        double f = 2.*atan2(sqrt((1.+e)/(1.-e*cos(E))) * sin(0.5*E),
                sqrt((1.-e)/(1-e*cos(E))) * cos(0.5*E));
        return TO_DEG(f);
    }
    return INVALID_NUMBER;
}

gsl_matrix* K_getElements(ok_kernel* k) {
    K_validate(k);
    return k->system->elements;
}
void K_setElements(ok_kernel* k, gsl_matrix* elements) {
    if (elements != k->system->elements)
        gsl_matrix_free(k->system->elements);
    
    k->system->nplanets = MROWS(elements) - 1;
    k->system->elements = elements;
    k->flags |= NEEDS_SETUP;
    K_validate(k);
}

gsl_matrix* K_getAllElements(ok_kernel* k) {
    K_validate(k);
    
    gsl_matrix* ret = ok_matrix_copy(k->system->elements);
    ret = ok_matrix_resize(ret, MROWS(k->system->elements), ALL_ELEMENTS_SIZE);
    
    for (int i = 0; i < MROWS(ret); i++)
        for (int j = ELEMENTS_SIZE; j < ALL_ELEMENTS_SIZE; j++) {
            MSET(ret, i, j, K_getElement(k, i, j));
        }
    
    return ret;
}

void K_setPars(ok_kernel* k, gsl_vector* pars) {
    if (pars != k->params)
        gsl_vector_free(k->params);
    k->params = ok_vector_resize(pars, PARAMS_SIZE);
}

gsl_vector* K_getPars(ok_kernel* k) {
    return k->params;
}

void K_setPar(ok_kernel* k, int idx, double val) {
    VSET(k->params, idx, val);
}

double K_getPar(ok_kernel* k, int idx) {
    return VGET(k->params, idx);
}

void K_setParFlag(ok_kernel* k, int idx, int value) {
    VSET(k->parFlags, idx, value);
}
int K_getParFlag(ok_kernel* k, int idx) {
    return VGET(k->parFlags, idx);
}


void K_setParRange(ok_kernel* k, int par, double min, double max) {
    if (k->parRanges[0] == NULL) {
        k->parRanges[0] = gsl_vector_alloc(PARAMS_SIZE);
        gsl_vector_set_all(k->parRanges[0], NAN);
    }
    if (k->parRanges[1] == NULL) {
        k->parRanges[1] = gsl_vector_alloc(PARAMS_SIZE);
        gsl_vector_set_all(k->parRanges[1], NAN);
    }
    K_validate(k);
    
    VSET(k->parRanges[0], par, min);
    VSET(k->parRanges[1], par, max);
}

void K_getParRange(ok_kernel* k, int idx, double* min, double* max) {
    K_validate(k);
    
    if (k->parRanges[0] == NULL)
        *min = INVALID_NUMBER;
    else
        *min = VGET(k->parRanges[0], idx);
    if (k->plRanges[1] == NULL)
        *max = INVALID_NUMBER;
    else
        *max = VGET(k->parRanges[1], idx);
}

void K_validate(ok_kernel* k) {
    
    if (k->flags & NEEDS_COMPILE) {
        K_compileData(k);
    }
    
    if (k->plFlags == NULL || (MROWS(k->plFlags) != MROWS(k->system->elements))) {
        k->plFlags = ok_matrix_int_resize(k->plFlags, MROWS(k->system->elements), MCOLS(k->system->elements));
    }
    
    if ((k->plSteps == NULL) || MROWS(k->plSteps) != MROWS(k->system->elements)) {
        k->plSteps = ok_matrix_resize(k->plSteps, MROWS(k->system->elements), MCOLS(k->system->elements));
    } 
    
    if (k->plRanges[0] != NULL && MROWS(k->plRanges[0]) != MROWS(k->system->elements)) {
        k->plRanges[0] = ok_matrix_resize_pad(k->plRanges[0], MROWS(k->system->elements), ELEMENTS_SIZE, INVALID_NUMBER);
        k->plRanges[1] = ok_matrix_resize_pad(k->plRanges[1], MROWS(k->system->elements), ELEMENTS_SIZE, INVALID_NUMBER);
    }

    if (k->parRanges[0] != NULL)
        for (int i = 0; i < PARAMS_SIZE; i++) {
            if (!IS_INVALID(VGET(k->parRanges[0], i)))
                VSET(k->params, i, MAX(VGET(k->parRanges[0], i), VGET(k->params, i)));
            if (!IS_INVALID(VGET(k->parRanges[1], i)))
                VSET(k->params, i, MIN(VGET(k->parRanges[1], i), VGET(k->params, i)));
        }
    
    for (int i = 1; i < k->system->nplanets+1; i++) {
        
        double ecc = MGET(k->system->elements, i, ECC);
        if (ecc < 0) {
            MSET(k->system->elements, i, ECC, -ecc);
            MINC(k->system->elements, i, MA, - 180.);
            MINC(k->system->elements, i, LOP, + 180.);
        }
        
        double mass = MGET(k->system->elements, i, MASS);
        if (mass < 0) {
            MSET(k->system->elements, i, MASS, -mass);
            MINC(k->system->elements, i, MA, -180.);
        }
        if (k->plRanges[0] != NULL) {
            for (int j = 0; j < ELEMENTS_SIZE; j++) {
                if (! IS_INVALID(MGET(k->plRanges[0], i, j)))
                    MSET(k->system->elements, i, j, MAX(MGET(k->system->elements, i, j), MGET(k->plRanges[0], i, j)));
                if (! IS_INVALID(MGET(k->plRanges[1], i, j)))
                    MSET(k->system->elements, i, j, MIN(MGET(k->system->elements, i, j), MGET(k->plRanges[1], i, j)));
            }       
        }
        
        
        double per = MGET(k->system->elements, i, PER);
        mass = MGET(k->system->elements, i, MASS);
        ecc = MGET(k->system->elements, i, ECC);
        
        
        
        double Mcent = 0.;
        
        for (int j = 0; j < k->system->nplanets + 1; j++)
            if (MGET(k->system->elements, j, PER) < per)
                Mcent += MGET(k->system->elements, j, MASS) * (j == 0 ? MSUN : MJUP);
        
        double a = cbrt(DAY_TO_YEAR(per) * DAY_TO_YEAR(per) * (Mcent + mass * MJUP) / MSUN);
        MSET(k->system->elements, i, SMA, a);
        
        // TODO : Add inclination
        double K = 2 * M_PI/per * mass * MJUP / Mcent * a / sqrt(1 - ecc*ecc);
        MSET(k->system->elements, i, SEMIAMP, K * AU / (100.*DAY));
    }

    MSET(k->system->elements, 0, MASS, k->Mstar);
    
    if (k->flags & NEEDS_SETUP)
        ok_setup(k->system);

    
}

int K_getActiveElements(ok_kernel* k) {
    int flags = 0;
    for (int i = 0; i < MROWS(k->plFlags); i++)
        for (int j = 0; j < MCOLS(k->plFlags); j++)
            if (MIGET(k->plFlags, i, j) & ACTIVE)
                flags++;

    return flags;
}

int K_getActivePars(ok_kernel* k) {
    int flags = 0;
    for (int i = 0; i < k->parFlags->size; i++)
        if (VIGET(k->parFlags, i) & ACTIVE)
            flags++;

    return flags;
}

void K_calculate(ok_kernel* k) {
    K_validate(k);

    int ppars = K_getActiveElements(k);
    int vpars = K_getActivePars(k);
    
    double pars = k->ndata - ppars - vpars;
    k->nrpars = pars;
    
    bool integrate = (k->system->nplanets > 0);
    
    if (k->ndata <= 0)
        return;
    
    if ((!integrate && k->integration != NULL) || ((k->integration != NULL) && (k->times->size != k->integrationSamples ||
            MROWS(k->integration[0]->elements) != MROWS(k->system->elements)))) {
        for (int i = 0; i < k->integrationSamples; i++)
            ok_free_system(k->integration[i]);
        free(k->integration);
        k->integration = NULL;
        k->integrationSamples = 0;
    }

    if (integrate) {
        k->integration = ok_integrate(k->system, k->times, k->intOptions, k->intMethod, k->integration);
    } else {
        for (int i = 0; i < k->ndata; i++) {
            k->compiled[i][T_PRED] = 0.;
        }
    }
    
    k->chi2_rvs = 0.;
    k->chi2_tts = 0.;
    
    k->rms = 0.;
    k->rms_tts = 0.;
    k->jitter = 0.;
    k->nrvs = 0;
    k->ntts = 0;
    
    for (int i = 0; i < k->ndata; i++) {
        if ((int) k->compiled[i][T_FLAG] == T_RV) {
            k->compiled[i][T_PRED] = (integrate ? ok_get_rv(k->integration[i]) : 0.0);
            int j = (int) k->compiled[i][T_SET];
            k->compiled[i][T_SVAL] = k->compiled[i][T_VAL] - VGET(k->params, j);
            double diff = k->compiled[i][T_SVAL] - k->compiled[i][T_PRED];
            double s = k->compiled[i][T_ERR];
            k->chi2_rvs += diff * diff / (s * s);
            k->rms += diff * diff;
            k->jitter += s*s;
            k->nrvs ++;
        } else if ((int) k->compiled[i][T_FLAG] == T_TIMING) {
            int pidx = (int) k->compiled[i][T_TDS_PLANET];
            if (pidx <= 0)
                pidx = 1;
            if (pidx >= k->system->nplanets + 1)
                continue;
            
            k->compiled[i][T_SVAL] = k->compiled[i][T_VAL];
            if (integrate) {
                ok_find_closest_transit(k->integration[i],
                    pidx, k->intOptions, k->intMethod, 1e-8, &(k->compiled[i][T_PRED]));
            } else {
                k->compiled[i][T_PRED] = 0.;
            }
            
            double diff = k->compiled[i][T_SVAL] - k->compiled[i][T_PRED];
            double s = k->compiled[i][T_ERR];
            k->chi2_tts += diff * diff / (s * s);
            k->rms_tts += diff;
            k->ntts ++;
        }
    }
    
    k->chi2 = (k->chi2_rvs + k->chi2_tts) / pars;
    k->rms = sqrt(k->rms / (double) k->nrvs);
    k->rms_tts = sqrt(k->rms_tts / (double) k->ntts);
    k->jitter = sqrt(k->rms * k->rms - k->jitter / (double) k->nrvs);
    
    k->integrationSamples = k->times->size;
    k->flags &= ~NEEDS_COMPILE;
    k->flags &= ~NEEDS_SETUP;
    
}

ok_system** K_integrate(ok_kernel* k, gsl_vector* times, ok_system** bag) {
    ok_setup(k->system);
    return ok_integrate(k->system, times, k->intOptions, k->intMethod, bag);
}

ok_system** K_integrateRange(ok_kernel* k, double from, double to, unsigned int samples, ok_system** bag) {
    gsl_vector* times = gsl_vector_alloc(samples);
    for (int i = 0; i < samples; i++)
        VSET(times, i, i * (to-from)/(samples-1) + from);
    ok_system** sl = K_integrate(k, times, bag);
    gsl_vector_free(times);
    return sl;
}

K_GETSET_C(Mstar, Mstar, double)
K_GETSETW_C(epoch, Epoch, double, k->system->epoch)
K_GET_C(chi2, Chi2, double)
K_GET_C(chi2_rvs, Chi2_rvs, double)
K_GET_C(chi2_tts, Chi2_tts, double)
K_GET_C(rms, Rms, double)
K_GET_C(rms_tts, Rms_tts, double)
K_GET_C(jitter, Jitter, double)
K_GETSET_C(minfunc, MinFunc, ok_callback)
K_GETSET_C(intMethod, IntMethod, int)
K_GETSET_C(intOptions, IntOptions, ok_integrator_options*)
K_GET_C(compiled, Compiled, double**)
K_GETSET_C(flags, Flags, unsigned int)

K_GETSET_FREE_C(plSteps, ElementSteps, gsl_matrix*, gsl_matrix_free)
K_GETSET_FREE_C(parSteps, ParSteps, gsl_vector*, gsl_vector_free)
K_GETSET_FREE_C(plFlags, ElementFlags, gsl_matrix_int*, gsl_matrix_int_free)
K_GETSET_FREE_C(parFlags, ParFlags, gsl_vector_int*, gsl_vector_int_free)


void K_setElementFlag(ok_kernel* k, int row, int col, int value) {
    K_validate(k);
    MISET(k->plFlags, row, col, value);
}

int K_getElementFlag(ok_kernel* k, int row, int col) {
    K_validate(k);
    return MIGET(k->plFlags, row, col);
}

void K_setVoffFlag(ok_kernel* k, int idx, int value) {
    K_validate(k);
    VISET(k->parFlags, idx, value);
}

int K_getVoffFlag(ok_kernel* k, int idx) {
    K_validate(k);
    return VIGET(k->parFlags, idx);
}

void K_setElementStep(ok_kernel* k, int row, int col, double value) {
    K_validate(k);
    
    if (row == -1 && col == -1 && IS_INVALID(value)) {
        for (int i = 1; i < MROWS(k->system->elements); i++) 
            for (int j = 0; j < ELEMENTS_SIZE; j++)
                MSET(k->plSteps, i, j, ok_default_steps[j]);
        return;
    }
    
    if (row == -1) {
        for (int i = 1; i < MROWS(k->system->elements); i++)
            MSET(k->plSteps, i, col, value);
    } else
        MSET(k->plSteps, row, col, value);
}

double K_getElementStep(ok_kernel* k, int row, int col) {
    K_validate(k);
    return MGET(k->plSteps, row, col);
}

void K_setElementRange(ok_kernel* k, int row, int col, double min, double max) {
    if (k->plRanges[0] == NULL) {
        k->plRanges[0] = gsl_matrix_alloc(MROWS(k->system->elements), ELEMENTS_SIZE);
        gsl_matrix_set_all(k->plRanges[0], NAN);
    }
    if (k->plRanges[1] == NULL) {
        k->plRanges[1] = gsl_matrix_alloc(MROWS(k->system->elements), ELEMENTS_SIZE);
        gsl_matrix_set_all(k->plRanges[1], NAN);
    }
    K_validate(k);
    
    if (row == -1) {
        for (int i = 1; i < MROWS(k->system->elements); i++) {
            MSET(k->plRanges[0], i, col, min);
            MSET(k->plRanges[1], i, col, max);
        }
    } else {
        MSET(k->plRanges[0], row, col, min);
        MSET(k->plRanges[1], row, col, max);
    }
}

void K_getElementRange(ok_kernel* k, int row, int col, double* min, double* max) {
    K_validate(k);
    
    if (k->plRanges[0] == NULL)
        *min = INVALID_NUMBER;
    else
        *min = MGET(k->plRanges[0], row, col);
    if (k->plRanges[1] == NULL)
        *max = INVALID_NUMBER;
    else
        *max = MGET(k->plRanges[1], row, col);
}

void K_setVoffStep(ok_kernel* k, int idx, double value) {
    K_validate(k);
    VSET(k->parSteps, idx, value);
}

double K_getVoffStep(ok_kernel* k, int idx) {
    K_validate(k);
    return VGET(k->parSteps, idx);
}

ok_kernel* K_clone(ok_kernel* k) {
    return K_cloneFlags(k, 0);
}

ok_kernel* K_cloneFlags(ok_kernel* k, unsigned int flags) {
    ok_kernel* k2 = (ok_kernel*) malloc(sizeof(ok_kernel));
    memcpy(k2, k, sizeof(ok_kernel));
    
    if (! (flags & SHARE_DATA)) {
        k2->datasets = (gsl_matrix**) malloc(sizeof(gsl_matrix*) * k->nsets);

        for (int i = 0; i < k->nsets; i++) {
            k2->datasets[i] = ok_matrix_copy(k->datasets[i]);
        }
        
        k2->compiled = NULL;
        k2->flags |= NEEDS_COMPILE;
        k2->times = NULL;
        k2->integration = NULL;
        k2->integrationSamples = -1;
        k2->plRanges = (gsl_matrix**) calloc(2, sizeof(gsl_matrix*));
        k2->parRanges = (gsl_vector**) calloc(2, sizeof(gsl_vector*));
    }
    
    if (! (flags & SHARE_STEPS)) {
        k2->plSteps = ok_matrix_copy(k->plSteps);
        k2->parSteps = ok_vector_copy(k->parSteps);
    }
    if (! (flags & SHARE_FLAGS)) {
        k2->plFlags = ok_matrix_int_copy(k->plFlags);
        k2->parFlags = ok_vector_int_copy(k->parFlags);
    }
    
    k2->system = ok_copy_system(k->system);
    k2->system->xyz = ok_matrix_copy(k->system->xyz);
    k2->system->orbits = ok_matrix_copy(k->system->orbits);
    
    k2->params = ok_vector_copy(k->params);
    
    k2->system->epoch = k->system->epoch;
    k2->system->time = k->system->time;
    k2->tag = k->tag;
    k2->rng = gsl_rng_alloc(gsl_rng_default);
    gsl_rng_set(k2->rng, clock());
    k2->flags |= flags; 
    k2->progress = k->progress;
    
    return k2;
}

int K_minimize(ok_kernel* k, int algo, int maxiter, double params[]) {
    int ret = ok_minimizers[algo](k, maxiter, params);
    
    for (int i = 1; i < k->system->elements->size1; i++) {
        MSET(k->system->elements, i, MA, DEGRANGE(MGET(k->system->elements, i, MA)));
        MSET(k->system->elements, i, LOP, DEGRANGE(MGET(k->system->elements, i, LOP)));
        MSET(k->system->elements, i, INC, DEGRANGE(MGET(k->system->elements, i, INC)));
        MSET(k->system->elements, i, NODE, DEGRANGE(MGET(k->system->elements, i, NODE)));
    }
    
    return ret;
}

int K_1dminimize(ok_kernel* k, int algo, int maxiter, int row, int column, double params[]) {
    gsl_matrix_int* m = ok_matrix_int_copy(k->plFlags);
    gsl_vector_int* p = ok_vector_int_copy(k->parFlags);
    
    for (int i = 0; i < m->size1; i++)
        for (int j = 0; j < m->size2; j++)
            MISET(m, i, j, MIGET(m, i, j) & ~MINIMIZE);
    for (int i = 0; i < p->size; i++)
        VISET(p, i, VIGET(p, i) & ~MINIMIZE);
    
    gsl_matrix_int* old_plFlags = k->plFlags;
    gsl_vector_int* old_parFlags = k->parFlags;
    
    if (row > 0) {
        MISET(m, row, column, ACTIVE | MINIMIZE);
    } else if (row == -1) {
        VISET(p, column, ACTIVE | MINIMIZE);
    }
    
    
    k->plFlags = m;
    k->parFlags = p;
    int ret = K_minimize(k, algo, maxiter, params);
    k->plFlags = old_plFlags;
    k->parFlags = old_parFlags;
    gsl_matrix_int_free(m);
    gsl_vector_int_free(p);
    return ret;
}

void K_print(ok_kernel* k, FILE* f) {
    f = (f == NULL ? stdout : f);
    fprintf(f, "Elements:\n");
    ok_fprintf_matrix(k->system->elements, f, "%e ");
    fprintf(f, "Data:\n");
    for (int i = 0; i < k->nsets; i++) {
        fprintf(f, "data[%d] -> %e\n", i, k->params->data[i]);
    }
    fprintf(f, "Orbits:\n");
    ok_fprintf_matrix(k->system->orbits, f, "%e ");
    fprintf(f, "XYZ:\n");
    ok_fprintf_matrix(k->system->xyz, f, "%e ");
    fprintf(f, "Compiled data:\n");
    ok_fprintf_buf(k->compiled, f, "%e ", k->ndata, DATA_SIZE);
    fflush(f);
}



void K_save_old(ok_kernel* k, const char* stem) {
    K_validate(k);
    
    
    char* fitName = ok_str_cat(stem, ".fit");
    char dataName[k->nsets][255];
    for (int i = 0; i < k->nsets; i++) {
        sprintf(dataName[i], "%s_%d.vels", stem, i);
    }
    char* sysName = ok_str_cat(stem, ".sys");
    
    FILE* fit = fopen(fitName, "w");
    fprintf(fit, "# InitialEpoch: %e\n\n", k->system->epoch);
    fprintf(fit, "Components %d\n", k->system->nplanets-1);
    fprintf(fit, "PrimaryRVSet \"%s\"\n", dataName[0]);
    fprintf(fit, "OverallRVOffset %e\n", VGET(k->params, 0));
    fprintf(fit, "RelativeRVOffsets {\n");
    for (int i = 1; i < k->nsets; i++)
        fprintf(fit, "\t\"%s\" %e\n", dataName[i], VGET(k->params, 1));
    fprintf(fit, "\t\"Trend\" %e\n}\n", 0.);
    
    for (int i = 1; i < k->system->nplanets + 1; i++) {
        fprintf(fit, "\"\" {\n");
        fprintf(fit, "\tPeriod %e\n", K_getElement(k, i, PER));
        fprintf(fit, "\tMass %e\n", K_getElement(k, i, MASS));
        fprintf(fit, "\tMeanAnomaly %e\n", K_getElement(k, i, MA));
        fprintf(fit, "\tEccentricity %e\n", K_getElement(k, i, ECC));
        fprintf(fit, "\tLongOfPericenter %e\n", K_getElement(k, i, LOP));
        fprintf(fit, "\tInclination %e\n", K_getElement(k, i, INC));
        fprintf(fit, "\tNode %e\n", K_getElement(k, i, NODE));
        fprintf(fit, "\tRadius %e\n", K_getElement(k, i, RADIUS));
        fprintf(fit, "}\n");
    }
    fclose(fit);
    
    for (int i = 0; i < k->nsets; i++) {
        FILE* vels = fopen(dataName[i], "w");
        ok_fprintf_buf(k->compiled, vels, "%e ", k->ndata, DATA_SIZE);
        fclose(vels);
    }
    
    FILE* sys = fopen(sysName, "w");
    fprintf(sys, "Data {\n");
    for (int i = 0; i < k->nsets; i++)
        fprintf(sys, "\tRV[] \"%s\"\n", dataName[i]);
    fprintf(sys, "}\n\"%s\" {\n Mass %e\n}\n", stem, k->Mstar);
    fclose(sys);
    free(fitName);
    free(sysName);
}


bool K_save(ok_kernel* k, FILE* fid) {
    int digits = 15;
    int fract = 10;
    char fmt[] = "%15.10e ";
    
    fid = (fid != NULL ? fid : stdout);
    fprintf(fid, "@Kernel\n\n");
    fprintf(fid, "Epoch = %*.*e\n", digits, fract, k->system->epoch);
    fprintf(fid, "Mstar = %*.*e\n", digits, fract, k->Mstar);
    fprintf(fid, "Chi2 = %*.*e\n", digits, fract, k->chi2);
    fprintf(fid, "RMS = %*.*e\n", digits, fract, k->rms);
    fprintf(fid, "Jitter = %*.*e\n", digits, fract, k->jitter);
    fprintf(fid, "IntMethod = %d\n", k->intMethod);
    
    fprintf(fid, "\nElements = %zu\n", k->system->elements->size1);
    ok_fprintf_matrix(k->system->elements, fid, fmt);
    
    fprintf(fid, "\nParams = \n");
    ok_fprintf_vector(k->params, fid, fmt);
    
    fprintf(fid, "\nPlFlags = %zu\n", k->plFlags->size1);
    ok_fprintf_matrix_int(k->plFlags, fid, "%d ");
    
    fprintf(fid, "\nParFlags = \n");
    ok_fprintf_vector_int(k->parFlags, fid, "%d ");
    
    fprintf(fid, "\nPlSteps = %zu\n", k->plSteps->size1);
    ok_fprintf_matrix(k->plSteps, fid, fmt);
    
    fprintf(fid, "\nParSteps = \n");
    ok_fprintf_vector(k->parSteps, fid, fmt);
    
    
    if (k->plRanges[0] != NULL) {
        fprintf(fid, "\nPlRanges_min = %zu\n", k->plRanges[0]->size1);
        ok_fprintf_matrix(k->plRanges[0], fid, fmt);
    }
    if (k->plRanges[1] != NULL) {
        fprintf(fid, "\nPlRanges_max = %zu\n", k->plRanges[1]->size1);
        ok_fprintf_matrix(k->plRanges[1], fid, fmt);
    }
    if (k->parRanges[0] != NULL) {
        fprintf(fid, "\nParRanges_min =\n");
        ok_fprintf_vector(k->parRanges[0], fid, fmt);
    }
    if (k->parRanges[1] != NULL) {
        fprintf(fid, "\nParRanges_max =\n");
        ok_fprintf_vector(k->parRanges[1], fid, fmt);
    }
    
    for (int i = 0; i < k->nsets; i++) {
        fprintf(fid, "\nData = %zu\n", k->datasets[i]->size1);
        ok_fprintf_matrix(k->datasets[i], fid, fmt);
    }
    
    fprintf(fid, "\n@End\n");
    return true;
}

ok_kernel* K_load(FILE* fid, int skip) {
    fid = (fid != NULL ? fid : stdin);
    char line[4096];
    for (int i = 0; i < skip; i++) {
        while (fgets(line, sizeof(line), fid)) 
            if (strcmp(line, "@End\n") == 0)
                break;
        if (feof(fid))
            return NULL;
    }
    
    bool found = false;
    char* ret;
    while (!found) {
       while ((ret = fgets(line, sizeof(line), fid)) != NULL) {
            if (strcmp(line, "@Kernel\n") == 0) {
                found = true;
                break;
            }
            if (feof(fid))
                return NULL;
       }

        if (ret == NULL) {
            return NULL;
        }
            
    }
    
    ok_kernel* k = K_alloc();
    
    
    double v;
    int r;
    k->nsets = 0;
    
    while (true) {
        if ((fgets(line, sizeof(line), fid) == NULL) || feof(fid) || strcmp(line, "@End\n") == 0)
            break;
        char tag[100] = {0};
        sscanf(line, "%s =", tag);
        if (strcmp(tag, "Epoch") == 0) {
            sscanf(line, "%*s = %le", &v);
            K_setEpoch(k, v);
        } else if (strcmp(tag, "Mstar") == 0) {
            sscanf(line, "%*s = %le", &v);
            K_setMstar(k, v);
        } else if (strcmp(tag, "IntMethod") == 0) {
            sscanf(line, "%*s = %d", &r);
            k->intMethod = r;
        } else if (strcmp(tag, "Elements") == 0) {
            sscanf(line, "%*s = %d", &r);
            gsl_matrix* els = gsl_matrix_alloc(r, ELEMENTS_SIZE);
            ok_fscanf_matrix(els, fid);
            gsl_matrix_free(k->system->elements);
            k->system->nplanets = r - 1;
            k->system->elements = els;
        } else if (strcmp(tag, "Params") == 0) {
            ok_fscanf_vector(k->params, fid);
        } else if (strcmp(tag, "PlSteps") == 0) {
            sscanf(line, "%*s = %d", &r);
            gsl_matrix* ps = gsl_matrix_alloc(r, ELEMENTS_SIZE);
            ok_fscanf_matrix(ps, fid);
            gsl_matrix_free(k->plSteps);
            k->plSteps = ps;
        } else if (strcmp(tag, "ParSteps") == 0) {
            ok_fscanf_vector(k->parSteps, fid);
        } else if (strcmp(tag, "PlFlags") == 0) {
            sscanf(line, "%*s = %d", &r);
            gsl_matrix_int* plflags = gsl_matrix_int_alloc(r, ELEMENTS_SIZE);
            ok_fscanf_matrix_int(plflags, fid);
            gsl_matrix_int_free(k->plFlags);
            k->plFlags = plflags;
        } else if (strcmp(tag, "ParFlags") == 0) {
            ok_fscanf_vector_int(k->parFlags, fid);
        } else if (strcmp(tag, "PlRanges_min") == 0) {
            sscanf(line, "%*s = %d", &r);
            gsl_matrix* pr = gsl_matrix_alloc(r, ELEMENTS_SIZE);
            ok_fscanf_matrix(pr, fid);
            k->plRanges[0] = pr;
        } else if (strcmp(tag, "PlRanges_max") == 0) {
            sscanf(line, "%*s = %d", &r);
            gsl_matrix* pr = gsl_matrix_alloc(r, ELEMENTS_SIZE);
            ok_fscanf_matrix(pr, fid);
            k->plRanges[1] = pr;
        } else if (strcmp(tag, "ParRanges_min") == 0) {
            gsl_vector* pr = gsl_vector_alloc(PARAMS_SIZE);
            ok_fscanf_vector(pr, fid);
            k->parRanges[0] = pr;
        } else if (strcmp(tag, "ParRanges_max") == 0) {
            gsl_vector* pr = gsl_vector_alloc(PARAMS_SIZE);
            ok_fscanf_vector(pr, fid);
            k->parRanges[1] = pr;
        } else if (strcmp(tag, "Data") == 0) {
            sscanf(line, "%*s = %d", &r);
            gsl_matrix* d = gsl_matrix_alloc(r, DATA_SIZE);
            ok_fscanf_matrix(d, fid);
            k->datasets[k->nsets] = d;
            k->nsets++;
        }
    }
    
    k->flags = NEEDS_SETUP | NEEDS_COMPILE;
    K_validate(k);
    
    return k;
    
}

void K_setSeed(ok_kernel* k, unsigned long int seed) {
    gsl_rng_set(k->rng, seed);
}


unsigned int K_getNplanets(ok_kernel* k) {
    return k->system->nplanets;
}
unsigned int K_getNsets(ok_kernel* k) {
    return k->nsets;
}
unsigned int K_getNdata(ok_kernel* k) {
    K_validate(k);
    return k->ndata;
}

void K_getRange(ok_kernel* k, double* from, double* to) {
    K_compileData(k);
    *from = k->compiled[0][0];
    *to = k->compiled[k->ndata-1][0];
}

gsl_matrix* K_combineData(ok_kernel* k, const int data_type) {
    gsl_matrix* filt;
    double** cd = K_compileData(k);
    
    if (data_type >= 0) 
        filt = ok_matrix_buf_filter(cd, k->ndata, DATA_SIZE, T_FLAG, data_type);
    else {
        filt = gsl_matrix_alloc(k->ndata, DATA_SIZE);
        
        for (int i = 0; i < k->ndata; i++)
            for (int j = 0; j < DATA_SIZE; j++)
                MSET(filt, i, j, cd[i][j]);
    }
    if (filt == NULL)
        return NULL;
    
    double* means = (double*) calloc(k->nsets, sizeof(double));
    double* n = (double*) calloc(k->nsets, sizeof(double));
    for (int i = 0; i < filt->size1; i++) {
        int set = (int) MGET(filt, i, T_SET);
        means[set] += MGET(filt, i, T_VAL);
        n[set] += 1.;
    }
    
    
    for (int i = 0; i < filt->size1; i++) {
        double val = MGET(filt, i, T_VAL);
        double pred = MGET(filt, i, T_PRED);
        int set = (int) MGET(filt, i, T_SET);
        MSET(filt, i, T_SVAL, val - means[set]/n[set]);
        MSET(filt, i, T_PRED, val - pred);
    }
    return filt;
}

