
#include "kl.h"
#include "kernel.h"
#include "gsl/gsl_vector.h"
#include "gsl/gsl_statistics.h"
#include "gsl/gsl_sort_vector.h"

ok_list* KL_alloc(int size) {
    ok_list* kl = (ok_list*) malloc(sizeof(ok_list));
    kl->kernels = (ok_kernel**) calloc(size, sizeof(ok_kernel*));
    kl->size = size;
    return kl;
}

void KL_free(ok_list* kl) {
    if (kl == NULL)
        return;
    
    for (int i = 0; i < kl->size; i++) {
        kl->kernels[i]->system->flag = FREED;
    }
        
    free(kl->kernels);
    free(kl);
}

gsl_vector* KL_getElements(ok_list* kl, int pl, int el) {
    gsl_vector* v = gsl_vector_alloc(kl->size);
    
    for (int i = 0; i < kl->size; i++)
        VSET(v, i, K_getElement(kl->kernels[i], pl, el));
    return v;
}

gsl_vector* KL_getPars(ok_list* kl, int vo) {
    gsl_vector* v = gsl_vector_alloc(kl->size);
    
    for (int i = 0; i < kl->size; i++)
        VSET(v, i, K_getPar(kl->kernels[i], vo));
    
    return v;
}

gsl_matrix* KL_getElementsStats(ok_list* kl, int what) {
    
    gsl_vector* v = gsl_vector_alloc(kl->size);
    int npl = MROWS(kl->kernels[0]->system->elements);
    gsl_matrix* m = gsl_matrix_alloc(npl, ELEMENTS_SIZE);
    gsl_matrix_set_all(m, 0.);
    
    for (int i = 0; i < npl; i++)
            for (int j = 0; j < ELEMENTS_SIZE; j++) {
                for (int n = 0; n < kl->size; n++) {
                    VSET(v, n, MGET(kl->kernels[n]->system->elements, i, j));
                }

                switch (what) {
                    case STAT_MEAN:
                        MSET(m, i, j, gsl_stats_mean(v->data, 1, v->size));
                        break;
                    case STAT_STDDEV:
                        MSET(m, i, j, gsl_stats_sd(v->data, 1, v->size));
                        break;
                    case STAT_MEDIAN:
                        gsl_sort_vector(v);
                        MSET(m, i, j, gsl_stats_median_from_sorted_data(v->data, 1, v->size));
                        break;
                    default:
                        // percentiles
                        gsl_sort_vector(v);
                        MSET(m, i, j, gsl_stats_quantile_from_sorted_data(v->data, 1, v->size, (double)(what)/100.));
                };
            }
    gsl_vector_free(v);
    return m;
}

unsigned int KL_getSize(ok_list* kl) {
    return kl->size;
}

void KL_removeAtIndex(ok_list* kl, unsigned int idx) {
    K_free(kl->kernels[idx]);
    for (int i = idx + 1; i < kl->size; i++)
        kl->kernels[i-1] = kl->kernels[i];
    kl->size--;
}

void KL_fprintf(ok_list* kl, FILE* out, char* fmt, char* lfmt) {
    lfmt = (lfmt != NULL ? lfmt : "%10s%d ");
    
    fprintf(out, "# ");
    
    int np = kl->kernels[0]->system->nplanets + 1;
    int vo = PARAMS_SIZE;
    
    for (int i = 0; i < ELEMENTS_SIZE; i++)
        for (int j = 1; j < np; j++)
            fprintf(out, lfmt, ok_orb_labels[i], j);
    for (int i = 0; i < vo; i++)
        fprintf(out, lfmt, "PARAM", i);
    fprintf(out, "\n");
    
    for (int m = 0; m < kl->size; m++) {
        ok_kernel* k = kl->kernels[m];
        for (int i = 0; i < ELEMENTS_SIZE; i++)
            for (int j = 1; j < np; j++)
                fprintf(out, fmt, MGET(k->system->elements, j, i));
        for (int i = 0; i < vo; i++)
            fprintf(out, fmt, VGET(k->params, i));
        fprintf(out, fmt, k->chi2_rvs);
        fprintf(out, "\n");
    }
    
}