#include "math.h"
#include "gsl/gsl_multimin.h"
#include "gsl/gsl_vector.h"
#include "utils.h"
#include "kernel.h"

typedef struct  {
    ok_kernel* k;
    double** pars;
} ok_simplex_params;

double K_simplex_f(const gsl_vector* cur, void* params) {
    ok_simplex_params* sp = (ok_simplex_params*) params;
    
    ok_kernel* k = sp->k;
    double** pars = sp->pars;
    
    for (int i = 0; i < cur->size; i++)
        *(pars[i]) = cur->data[i];
    
    k->flags |= NEEDS_SETUP;
    K_calculate(k);
    double res = k->minfunc(k);
    
    return res;
}

int K_minimize_simplex(ok_kernel* k, int maxiter, double params[]) {
    double eps = 1e-6;
    
    int i = 0;
    if (params != NULL) {
        while (true) {
            if (params[i] == DONE)
                break;
            else if (round(params[i]) == OPT_SIMPLEX_EPS)
                eps = params[i+1];
            i+=2;
        }
    }
    int npars = 0;
    
    // Count all the parameters to minimize on
    for (int i = 1; i < k->system->nplanets + 1; i++)
        for (int j = 0; j < ELEMENTS_SIZE; j++)
            npars += (MIGET(k->plFlags, i, j) & MINIMIZE ? 1 : 0);
    for (int i = 0; i < k->parFlags->size; i++)
        npars += (VIGET(k->parFlags, i) & MINIMIZE ? 1 : 0);

    // Create a pointer table (flat array -> matrices)
    double** pars = (double**) malloc(sizeof(double*) * npars);
    
    int idx = 0;
    gsl_vector* x = gsl_vector_alloc(npars);
    gsl_vector* step = gsl_vector_alloc(npars);
    
    for (int i = 1; i < k->system->nplanets + 1; i++)
        for (int j = 0; j < ELEMENTS_SIZE; j++)
            if (MIGET(k->plFlags, i, j) & MINIMIZE) {
                
                pars[idx] = gsl_matrix_ptr(k->system->elements, i, j);
                
                x->data[idx] = MGET(k->system->elements, i, j);
                step->data[idx] = MGET(k->plSteps, i, j);
                idx++;
            }

    for (int i = 0; i < k->parFlags->size; i++)
        if (VIGET(k->parFlags, i) & MINIMIZE) {
            pars[idx] = gsl_vector_ptr(k->params, i);
            x->data[idx] = VGET(k->params, i);
            step->data[idx] = VGET(k->parSteps, i);
            idx++;
        }
        
    ok_simplex_params sp;
    sp.k = k;
    sp.pars = pars;
    
    // Initialize minimizer
    gsl_multimin_fminimizer* s = gsl_multimin_fminimizer_alloc(gsl_multimin_fminimizer_nmsimplex2,
                npars);
    gsl_multimin_function f;
    f.n = npars;
    f.f = K_simplex_f;
    f.params = &sp;
    
    gsl_multimin_fminimizer_set(s, &f, x, step);
    
    
    int iter = 0;
    int status;
    
    // Loop until the desired eps (the size of the simplex) is reached.
    
    ok_progress pr = k->progress;
    
    do {
        iter++;
        status = gsl_multimin_fminimizer_iterate(s);
        if (status)
            break;
        double size = gsl_multimin_fminimizer_size(s);
        status = gsl_multimin_test_size(size, eps);
        if (status == GSL_SUCCESS)
            break;
        
        if (pr != NULL)
            if (pr(iter, maxiter, k, __func__) != PROGRESS_CONTINUE) {
                status = PROGRESS_STOP;
                break;
            }
    } while (iter < maxiter);
    
    gsl_multimin_fminimizer_free(s);
    free(pars);
    gsl_vector_free(step);
    gsl_vector_free(x);
    return status;
}
