/*
 *  types.h
 *  Systemic2
 *
 *
 */

#include "stdlib.h"
#include "stdbool.h"
#include "stdio.h"
#include "string.h"
#include "float.h"

#include "gsl/gsl_math.h"
#include "gsl/gsl_matrix.h"
#include "gsl/gsl_vector.h"
#include "gsl/gsl_rng.h"

#ifndef types_h
#define types_h

#define INVALID_NUMBER (NAN)
#define IS_INVALID(x) (isnan(x))

#define VERSION 1.1011

#define T_RV 0
#define T_PHOTO 1
#define T_TIMING 2

// Number of columns for data matrices. Columns 0-5 can be used to store data,
// while columns 6-10 are used internally.
#define DATA_SIZE 11
// Columns of data matrices
// Time (usually JD)
#define T_TIME 0
// Value (e.g. RV, flux, transit timing, etc.)
#define T_VAL 1
// Error
#define T_ERR 2
// Timing datasets: flag to associate a transit to a specific planet number. Set to 1 (first planet) by default.
#define T_TDS_PLANET 3
// Type of data
#define T_FLAG 6
// Shifted value (T_VAL shifted by the RV offset, otherwise same as T_VAL)
#define T_SVAL 7
// Computed value
#define T_PRED 8
// Dataset number
#define T_SET 9
// Reserved for GUI to do its dirty stuff
#define T_GUI 10

#define X 1
#define Y 2
#define Z 3
#define VX 4
#define VY 5
#define VZ 6

#define STAT_MEAN 0
#define STAT_MEDIAN 1
#define STAT_STDDEV 2
#define STAT_QUART_25 25
#define STAT_QUART_50 50
#define STAT_QUART_75 75

#define ENSURE_SIZE(matrix, rows, columns) assert((matrix->size1 == rows) && (matrix->size2 == columns))

#define KEPLER 0
#define RK5 1
#define RK8 2
#define SWIFT_BS 3
#define SWIFT_WHM 4
#define SWIFT_RMVS3 5

//#define K2 2.959122082855911e-4
#define AU  1.4959787e13
#define MSUN 1.98892e33
#define MJUP 1.8986e30
#define RJUP 7.1e9
#define GGRAV 6.67384e-8

#define DAY 8.64e4
#define TWOPI (2.*M_PI)
#define YEAR 31556926.

#define K2  ((GGRAV * MSUN * DAY * DAY) / (AU*AU*AU))

#define PER 0
#define MASS 1
#define MA 2
#define ECC 3
#define LOP 4
#define INC 5
#define NODE 6
#define ORD 14
#define RADIUS 7
#define SMA 8
#define SEMIAMP 9
// Extra elements that are calculated on-the-fly
#define ECCANOMALY 15
#define TRUEANOMALY 16

#define ELEMENTS_SIZE 15
#define ALL_ELEMENTS_SIZE 17

#define TREND 8
#define NOISE 9
#define PARAMS_SIZE 15

extern char * ok_orb_labels[15];

#define OPT_SIMPLEX_EPS 0
#define OPT_ECC_LAST 1

#define PROGRESS_CONTINUE 0
#define PROGRESS_STOP 1

#define DONE -1

typedef struct ok_system {
    /// The number of planets
    int nplanets;
    /// Initial epoch, to which the elements
    /// matrix below refers
    double epoch;
    /// Initial elements at epoch, in jacobian 
    /// coordinates.
    /// Units: period in days, mass in Mjup,
    /// MA, inc, node, lop in degrees.
    gsl_matrix* elements;

    /// The time to which the coordinates below 
    /// refer.
    double time;
    /// Cartesian coordinates. Internal units 
    /// are used.
    gsl_matrix* xyz;
    /// Current orbital elements. Internal units
    /// are used.
    gsl_matrix* orbits;
    /// flags
    int flag;
} ok_system;

typedef struct ok_integrator_options {
    double abs_acc;
    double rel_acc;
    /// A pointer to a force function that takes in time,
    /// the current state as an array, the output force
    /// array and a pointer to the current system state.
    int (* force) (double, const double[], double[], void*);
    int (* jac) (double, const double[], double*, double[], void*);
    // A generic pointer to a structure. Depending on the 
    // integrator, it may be useful to recycle memory.
    double** buffer;

} ok_integrator_options;

typedef struct ok_kernel ok_kernel;
typedef double(*ok_callback)(ok_kernel*);
typedef ok_system**(*ok_integrator)(ok_system*, const gsl_vector*, const ok_integrator_options*);
typedef int(*ok_minimizer)(ok_kernel*, int, double[]);
typedef int(*ok_progress)(int current, int max, void* state, const char* function);

struct ok_kernel {
    // initial conditions
    ok_system* system;
    // reduced chi^2
    double chi2;
    // chi^2
    double chi2_rvs;
    // chi^2 tts
    double chi2_tts;
    // rms of residuals
    double rms;
    // rms of tts
    double rms_tts;
    // jitter
    double jitter;
    // number of radial velocities
    int nsets;
    // list of radial velocities
    gsl_matrix** datasets;

    // compiled data (merged and sorted); each row is a pointer to a row
    // of data (RV, photometry, transit timing, etc.)
    double** compiled;
    // velocity offsets
    gsl_vector* params;
    // times
    gsl_vector* times;
    // result of last integration
    ok_system** integration;
    // number of integration samples
    int integrationSamples;

    // minimization flags
    gsl_matrix_int* plFlags;
    gsl_vector_int* parFlags;
    gsl_matrix* plSteps;
    gsl_vector* parSteps;
    gsl_matrix** plRanges;
    gsl_vector** parRanges;

    // stellar parameters
    double Mstar;

    // total number of data points
    int ndata;
    int nrvs;
    int ntts;
    int nrpars;
    
    // integration
    int intMethod;
    ok_integrator_options* intOptions;

    // function to minimize; usually, chi^2
    ok_callback minfunc;

    // flags
    unsigned int flags;

    // private random number generator
    gsl_rng* rng;

    // tagging data
    void* tag;
    
    // progress callback
    ok_progress progress;
};

typedef struct ok_list ok_list;

struct ok_list {
    ok_kernel** kernels;
    int size;
};

#endif
