//
//  utils.c
//  Systemic Console
//
//

#include <gsl/gsl_matrix_double.h>

#include "utils.h"
#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "systemic.h"
#include "assert.h"

double DEGRANGE(double angle) {
    return (fmod((angle < 0. ? angle + (floor(-angle/360.) + 1.) * 360. : angle), 360.));
}

double RADRANGE(double angle) {
    return (fmod((angle < 0. ? angle + (floor(-angle/TWOPI) + 1.) * TWOPI : angle), TWOPI));
}





/*  */
void ok_sort_small_matrix(gsl_matrix* matrix, const int column) {
    const int nrows = matrix->size1;
    int swaps = 1;
    while (swaps > 0) {
        swaps = 0;
        for (int i = 1; i < nrows; i++) {
            if (MGET(matrix, i, column) < MGET(matrix, i-1, column)) {
                gsl_matrix_swap_rows(matrix, i, i-1);
                swaps++;
            }
        }
    }
}

void ok_bootstrap_matrix(const gsl_matrix* matrix, gsl_matrix* out, const int sortcol, gsl_rng* r) {
    if (!out)
        out = gsl_matrix_alloc(matrix->size1, matrix->size2);
    
    for (int i = 0; i < matrix->size1; i++) {
        int k = gsl_rng_uniform_int(r, matrix->size1);
        
        for (int j = 0; j < matrix->size2; j++) {
            MSET(out, i, j, MGET(matrix, k, j));
        }
    }
    
    if (sortcol >= 0)
        ok_sort_matrix(out, sortcol);
}

void ok_bootstrap_matrix_mean(const gsl_matrix* matrix, int timecol, int valcol, gsl_matrix* out, gsl_rng* r) {
    if (!out)
        out = gsl_matrix_alloc(matrix->size1, matrix->size2);
    
    double mean = 0.;
    //for (int i = 0; i < matrix->size1; i++)
      //  mean += MGET(matrix, i, valcol);
    //mean /= (double) matrix->size1;
    
    for (int i = 0; i < matrix->size1; i++) {
        int k = gsl_rng_uniform_int(r, matrix->size1);
        
        for (int j = 0; j < matrix->size2; j++) {
            if (j == timecol) 
                MSET(out, i, j, MGET(matrix, i, j));
            else if (j == valcol) {
                MSET(out, i, j, MGET(matrix, k, j) - mean);
            } else 
                MSET(out, i, j, MGET(matrix, k, j));
        }
    }
    
}



void ok_fprintf_matrix(gsl_matrix* matrix, FILE* file, const char* format) {
    if (matrix == NULL) return;
    file = (file == NULL ? stdout : file);
    for (int i = 0; i < matrix->size1; i++) {
        for (int j = 0; j < matrix->size2; j++)
            fprintf(file, format, MGET(matrix, i, j));
        fprintf(file, "\n");
    }
}

void ok_fprintf_matrix_int(gsl_matrix_int* matrix, FILE* file, const char* format) {
    if (matrix == NULL) return;
    file = (file == NULL ? stdout : file);
    for (int i = 0; i < matrix->size1; i++) {
        for (int j = 0; j < matrix->size2; j++)
            fprintf(file, format, MIGET(matrix, i, j));
        fprintf(file, "\n");
    }
}

bool ok_save_matrix(gsl_matrix* matrix, FILE* fid, const char* format) {
    fid = (fid == NULL ? stdout : fid);
    ok_fprintf_matrix(matrix, fid, format);
    return true;
}

bool ok_save_matrix_bin(gsl_matrix* matrix, FILE* fid) {
    fid = (fid == NULL ? stdout : fid);
    gsl_matrix_fwrite(fid, matrix);
    return true;
}


bool ok_save_buf(double** matrix, FILE* fid, const char* format, int rows, int cols) {
    fid = (fid == NULL ? stdout : fid);
    ok_fprintf_buf(matrix, fid, format, rows, cols);
    return true;
}

bool ok_save_buf_bin(double** matrix, FILE* fid, int rows, int cols) {
    fid = (fid == NULL ? stdout : fid);
    for (int i = 0; i < rows; i++)
        fwrite(matrix[i], sizeof(double), cols, fid);
    return true;
}

void ok_fprintf_vector(gsl_vector* v, FILE* file, const char* format) {
    if (v == NULL) return;
    file = (file == NULL ? stdout : file);
    for (int i = 0; i < v->size; i++) 
       fprintf(file, format, VGET(v, i));
    fprintf(file, "\n");
}

void ok_fprintf_vector_int(gsl_vector_int* v, FILE* file, const char* format) {
    if (v == NULL) return;
    file = (file == NULL ? stdout : file);
    for (int i = 0; i < v->size; i++) 
       fprintf(file, format, VIGET(v, i));
    fprintf(file, "\n");
}

void ok_fprintf_buf(double** buf, FILE* file, const char* format, int rows, int columns) {
    if (buf == NULL) return;
    file = (file == NULL ? stdout : file);
    for (int i = 0; i < rows; i++) {
        for (int j = 0; j < columns; j++)
            fprintf(file, format, buf[i][j]);
        fprintf(file, "\n");
    }
}

gsl_vector* ok_vector_resize(gsl_vector* v, int len) {
    gsl_vector* nv = gsl_vector_calloc(len);
    if (v != NULL) {
        for (int i = 0; i < MIN(len, v->size); i++)
                VSET(nv, i, VGET(v, i));
        gsl_vector_free(v);
    }
    return nv;
}

gsl_vector* ok_vector_resize_pad(gsl_vector* v, int len, double pad) {
    gsl_vector* nv = gsl_vector_calloc(len);
    gsl_vector_set_all(v, pad);
    if (v != NULL) {
        for (int i = 0; i < MIN(len, v->size); i++)
                VSET(nv, i, VGET(v, i));
        gsl_vector_free(v);
    }
    return nv;
}

gsl_vector_int* ok_vector_int_resize(gsl_vector_int* v, int len) {
    gsl_vector_int* nv = gsl_vector_int_calloc(len);
    if (v != NULL) {
        for (int i = 0; i < MIN(len, v->size); i++)
                VISET(nv, i, VIGET(v, i));
        gsl_vector_int_free(v);
    }
    return nv;
}

gsl_matrix* ok_matrix_resize(gsl_matrix* v, int rows, int cols) {
    gsl_matrix* m = gsl_matrix_calloc(rows, cols);
    if (v != NULL) {
        for (int i = 0; i < MIN(rows, MROWS(v)); i++)
            for (int j = 0; j < MIN(cols, MCOLS(v)); j++)
                MSET(m, i, j, MGET(v, i, j));
        gsl_matrix_free(v);
    }
    return m;
}

gsl_matrix* ok_matrix_resize_pad(gsl_matrix* v, int rows, int cols, double pad) {
    gsl_matrix* m = gsl_matrix_calloc(rows, cols);
    gsl_matrix_set_all(m, pad);
    if (v != NULL) {
        for (int i = 0; i < MIN(rows, MROWS(v)); i++)
            for (int j = 0; j < MIN(cols, MCOLS(v)); j++)
                MSET(m, i, j, MGET(v, i, j));
        gsl_matrix_free(v);
    }
    return m;
}

gsl_matrix_int* ok_matrix_int_resize(gsl_matrix_int* v, int rows, int cols) {
    gsl_matrix_int* m = gsl_matrix_int_calloc(rows, cols);
    if (v != NULL) {
        for (int i = 0; i < MIN(rows, MROWS(v)); i++)
            for (int j = 0; j < MIN(cols, MCOLS(v)); j++)
                MISET(m, i, j, MIGET(v, i, j));
        gsl_matrix_int_free(v);
    }
    return m;
}

gsl_matrix* ok_read_table(FILE* fid) {
	const int MAXLEN = 4000;
	char line[MAXLEN];
    
    double JD = 0.0;
    int tds_plFlag = 1;
    
	fgets(line, MAXLEN, fid);
	while (line[0] == '#') {
        char tag[100] = {0};
        sscanf(&line[1], "%s =", tag);
        
        if (strcmp(tag, "JD") == 0) {
            sscanf(&line[1], "%*s = %le", &JD);
        } else if (strcmp(tag, "Planet") == 0) {
            sscanf(&line[1], "%*s = %d", &tds_plFlag);
        }
        
		fgets(line, MAXLEN, fid);
    }
		
	int nrows = 1;
        
	while (fgets(line, MAXLEN, fid) != NULL) {
		if (line[0] != '#')
			nrows++;
	}
	
	fseek(fid, 0, SEEK_SET); 
	
	gsl_matrix* ret = gsl_matrix_calloc(nrows, DATA_SIZE);
        double v[6];
	int nr = 0;
	while (fgets(line, MAXLEN, fid)!=NULL) {
		if (line[0] == '#')
			continue;
        for (int j = 0; j < 6; j++)
            v[j] = 0.;

        sscanf(line, "%le %le %le %le %le %le",
                &(v[0]), &(v[1]), &(v[2]),
                &(v[3]), &(v[4]), &(v[5]));
        for (int j = 0; j < 6; j++)
            MSET(ret, nr, j, v[j]);

		nr++;
	}
    
    for (int i = 0; i < nrows; i++) {
        MINC(ret, i, T_TIME, JD);
        if (tds_plFlag > 0)
            MSET(ret, i, T_TDS_PLANET, tds_plFlag);
    }
	
	return ret;
}

gsl_matrix* ok_matrix_copy(const gsl_matrix* src) {
    if (src == NULL) return NULL;
    gsl_matrix* m = gsl_matrix_alloc(MROWS(src), MCOLS(src));
    gsl_matrix_memcpy(m, src);
    return m;
}
gsl_matrix_int* ok_matrix_int_copy(const gsl_matrix_int* src) {
    if (src == NULL) return NULL;
    gsl_matrix_int* m = gsl_matrix_int_alloc(MROWS(src), MCOLS(src));
    gsl_matrix_int_memcpy(m, src);
    return m;
}
gsl_vector* ok_vector_copy(const gsl_vector* src) {
    if (src == NULL) return NULL;
    gsl_vector* v = gsl_vector_alloc(src->size);
    gsl_vector_memcpy(v, src);
    return v;
}
gsl_vector_int* ok_vector_int_copy(const gsl_vector_int* src) {
    if (src == NULL) return NULL;
    gsl_vector_int* v = gsl_vector_int_alloc(src->size);
    gsl_vector_int_memcpy(v, src);
    return v;
}

char* ok_str_copy(const char* src) {
    if (src == NULL) return NULL;
    char* d = (char*) malloc(strlen(src) * sizeof(char*));
    strcpy(d, src);
    return d;
}

char* ok_str_cat(const char* a1, const char* a2) {
    if (a1 == NULL) return ok_str_copy(a2);
    else if (a2 == NULL) return ok_str_copy(a1);
    
    char* a3 = (char*) malloc((strlen(a1) + strlen(a2) + 1) * sizeof(char));
    strcpy(a3, a1);
    strcat(a3, a2);
    return a3;
}

int ok_bsearch(double* v, double val, int len) {
    
    int min = 0;
    int max = len;
    
    while (max >= min) {
        int mid = (int)(0.5 * (min + max));
        if (v[mid] < val)
            min = mid + 1;
        else if (v[mid] > val)
            max = mid - 1;
        else if (v[mid] == val)
            return mid;
    }
    
    return min;
}

void ok_avevar(const double* v, int len, double* ave, double* var) {
    *ave = 0.;
    for (int i = 0; i < len; i++)
        *ave += v[i];
    *ave /= (double)len;
    *var = 0.;
    double sum2 = 0.;
    
    for (int i = 0; i < len; i++) {
        *var += SQR(v[i] - *ave);
        sum2 += v[i]-*ave;
    }
    
    *var = (*var - sum2/(double) len) / (double)(len-1);
}

int _ok_compare(void* col, const void* row1, const void* row2) {
    int c = *((int*)col);
    double* d1 = (double*) row1;
    double* d2 = (double*) row2;
    
    if (d1[c] < d2[c])
        return -1;
    else if (d1[c] > d2[c]) 
        return 1;
    else
        return 0;
}

void ok_sort_matrix(gsl_matrix* matrix, const int column) {
    ok_qsort_r(matrix->data, matrix->size1, matrix->size2 * sizeof(double*), (void*)&column, _ok_compare);
}

int _ok_rcompare(void* col, const void* row1, const void* row2) {
    int c = *((int*)col);
    double* d1 = (double*) row1;
    double* d2 = (double*) row2;
    
    if (d1[c] < d2[c])
        return 1;
    else if (d1[c] > d2[c]) 
        return -1;
    else
        return 0;
}

void ok_rsort_matrix(gsl_matrix* matrix, const int column) {
    ok_qsort_r(matrix->data, matrix->size1, matrix->size2 * sizeof(double*), (void*)&column, _ok_compare);
}

gsl_matrix* ok_matrix_filter(gsl_matrix* matrix, const int column, const double filter) {
    int rows = 0;
    for (int i = 0; i < matrix->size1; i++)
        if (fabs(MGET(matrix, i, column) - filter) < 1e-10)
            rows++;
    gsl_matrix* ret = gsl_matrix_alloc(rows, matrix->size2);
    
    rows = 0;
    for (int i = 0; i < matrix->size1; i++)
        if (fabs(MGET(matrix, i, column) - filter) < 1e-10) {
            rows++;
            for (int j = 0; j < matrix->size2; j++)
                MSET(ret, rows, j, MGET(matrix, i, j));
        }
    
    return ret;
    
}

gsl_matrix* ok_matrix_buf_filter(double** matrix, const int rows, const int columns, const int column, const double filter) {
    int nrows = 0;
    for (int i = 0; i < rows; i++) 
        if (fabs(matrix[i][column] - filter) < 1e-10) 
            nrows++;
 
    gsl_matrix* ret = gsl_matrix_alloc(nrows, columns);
    
    nrows = 0;
    for (int i = 0; i < rows; i++) {
        if (fabs(matrix[i][column] - filter) < 1e-10) {
            for (int j = 0; j < columns; j++)
                MSET(ret, nrows, j, matrix[i][j]);
            nrows++;
        }
    }
    return ret;
    
}

void ok_fscanf_matrix(gsl_matrix* matrix, FILE* file) {
    double* data = matrix->data;
    for (int i = 0; i < matrix->size1; i++)
        for (int j = 0; j < matrix->size2; j++)
            fscanf(file, "%le", &(data[i*matrix->size2 + j]));
}
void ok_fscanf_matrix_int(gsl_matrix_int* matrix, FILE* file) {
    int* data = matrix->data;
    for (int i = 0; i < matrix->size1; i++)
        for (int j = 0; j < matrix->size2; j++)
            fscanf(file, "%d", &(data[i*matrix->size2 + j]));
}
void ok_fscanf_vector(gsl_vector* vector, FILE* file) {
    double* data = vector->data;
    for (int i = 0; i < vector->size; i++)
        fscanf(file, "%le", &(data[i]));
}
void ok_fscanf_vector_int(gsl_vector_int* vector, FILE* file) {
    int* data = vector->data;
    for (int i = 0; i < vector->size; i++)
        fscanf(file, "%d", &(data[i]));
}


gsl_matrix* ok_buf_to_matrix(double** buf, int rows, int cols) {
    gsl_matrix* m = gsl_matrix_alloc(rows, cols);
    for (int i = 0; i < rows; i++)
        for (int j = 0; j < cols; j++)
            MSET(m, i, j, buf[i][j]);
    return m;
}

gsl_matrix* ok_matrix_remove_row(gsl_matrix* m, int row) {
    gsl_matrix* n = gsl_matrix_alloc(m->size1-1, m->size2);
    
    for (int i = 0; i < m->size1; i++)
        for (int j = 0; j < m->size2; j++) {
            if (i < row)
                MSET(n, i, j, MGET(m, i, j));
            else if (i > row)
                MSET(n, i - 1, j, MGET(m, i, j));
        }
    
    return n;
}
gsl_matrix* ok_matrix_remove_column(gsl_matrix* m, int col) {
    gsl_matrix* n = gsl_matrix_alloc(m->size1-1, m->size2);
    
    for (int i = 0; i < m->size1; i++)
        for (int j = 0; j < m->size2; j++) {
            if (j < col)
                MSET(n, i, j, MGET(m, i, j));
            else if (i > col)
                MSET(n, i, j-1, MGET(m, i, j));
        }
    
    return n;
}

gsl_matrix_int* ok_matrix_int_remove_row(gsl_matrix_int* m, int row) {
    gsl_matrix_int* n = gsl_matrix_int_alloc(m->size1-1, m->size2);
    
    for (int i = 0; i < m->size1; i++)
        for (int j = 0; j < m->size2; j++) {
            if (i < row)
                MISET(n, i, j, MIGET(m, i, j));
            else if (i > row)
                MISET(n, i - 1, j, MIGET(m, i, j));
        }
    
    return n;
}

gsl_matrix_int* ok_matrix_int_remove_column(gsl_matrix_int* m, int col) {
    gsl_matrix_int* n = gsl_matrix_int_alloc(m->size1-1, m->size2);
    
    for (int i = 0; i < m->size1; i++)
        for (int j = 0; j < m->size2; j++) {
            if (j < col)
                MISET(n, i, j, MIGET(m, i, j));
            else if (i > col)
                MISET(n, i, j-1, MIGET(m, i, j));
        }
    
    return n;
}

gsl_vector* ok_vector_remove(gsl_vector* m, int idx) {
    gsl_vector* n = gsl_vector_alloc(m->size);
    
    for (int i = 0; i < m->size; i++) {
        if (i < idx)
            VSET(n, i, VGET(m, i));
        else if (i > idx)
            VSET(n, i - 1, VGET(m, i));
    }
    
    return n;
}

gsl_vector_int* ok_vector_int_remove(gsl_vector_int* m, int idx) {
    gsl_vector_int* n = gsl_vector_int_alloc(m->size);
    
    for (int i = 0; i < m->size; i++) {
        if (i < idx)
            VISET(n, i, VIGET(m, i));
        else if (i > idx)
            VISET(n, i - 1, VIGET(m, i));
    }
    
    return n;
}
