//
//  utils.h
//  Systemic Console

//

#include "gsl/gsl_matrix.h"
#include "gsl/gsl_rng.h"
#include "math.h"
#include "stdbool.h"

#ifndef UTILS_H
#define UTILS_H


#define DEBUG 

#ifdef __APPLE__
#define ok_qsort_r qsort_r
#else
typedef int		 cmp_t(void *, const void *, const void *);
extern void ok_qsort_r(void *a, size_t n, size_t es, void *thunk, cmp_t *cmp);
#endif

#define MIN(a,b) (a < b ? a : b)
#define MAX(a,b) (a > b ? a : b)

#define RANGE(x, a, b) (MIN(MAX(x, a), b))

#define MGET(m, i, j) (m->data[(i)*m->tda+(j)])
#define MSET(m, i, j, x) m->data[(i)*m->tda+(j)] = x 
#define MIGET gsl_matrix_int_get
#define MISET gsl_matrix_int_set

#define MINC(m, i, j, c) MSET(m, i, j, MGET(m, i, j) + c)
#define MROWS(m) (m->size1)
#define MCOLS(m) (m->size2)

#define SQR(x) ((x)*(x))

#define VGET(v, i) (v->data[(i)])
#define VSET(v, i, x) v->data[(i)] = (x)
#define VIGET gsl_vector_int_get
#define VISET gsl_vector_int_set
#define ROW gsl_matrix_row
#define COL gsl_matrix_col

#define VLEN(X) (sqrt(X[0] * X[0] + X[1] * X[1] + X[2] * X[2]))
#define DELTA(I, J) (I == J ? 1 : 0)

#define MSUN_TO_INT(M) (M * K2)
#define INT_TO_MSUN(M) (M / K2)

#define MJUP_TO_INT(M) (M * MJUP / MSUN * K2)
#define INT_TO_MJUP(M) (M / K2 * MSUN / MJUP)

#define MSUN_TO_MJUP(M) (M * MJUP / MSUN)
#define MJUP_TO_MSUN(M) (M * MSUN / MJUP)

#define GM_TO_INT(M) (M / MSUN * K2)
#define INT_TO_GM(M) (M / K2 * MSUN)

#define DAY_TO_YEAR(T) (T * DAY / YEAR)
#define YEAR_TO_DAY(T) (T / DAY * YEAR)

#define CM_TO_INT(L) (L / AU)
#define INT_TO_CM(L) (L * AU)

#define AUPDAY_TO_MPS(V) (V * AU / 100 / DAY)
#define MPS_TO_AUPDAY(V) (V * DAY / (AU / 100))

#define sqr(x) ((x)*(x))
#define TO_DEG(ANGLE) (DEGRANGE(ANGLE / TWOPI * 360.))
#define TO_RAD(ANGLE) (RADRANGE(ANGLE / 360. * TWOPI))



#define K_GETSET_H(name, Name, type) \
        void K_set##Name(ok_kernel* k, type name);\
        type K_get##Name(ok_kernel* k);

#define K_GETSET_C(name, Name, type) \
        void K_set##Name(ok_kernel* k, type name) { k-> name = name; K_validate(k); k->flags |= NEEDS_SETUP; } \
        type K_get##Name(ok_kernel* k) { return k-> name; }

#define K_GETSET_FREE_C(name, Name, type, freefcn) \
        void K_set##Name(ok_kernel* k, type name) { if (k-> name == name) return; freefcn(k->name); k-> name = name;  } \
        type K_get##Name(ok_kernel* k) { return k-> name; }


#define K_GETSETW_C(name, Name, type, where) \
        void K_set##Name(ok_kernel* k, type name) { where = name; } \
        type K_get##Name(ok_kernel* k) { return where; }

#define K_GET_H(name, Name, type) \
        type K_get##Name(ok_kernel* k);

#define K_GET_C(name, Name, type) \
        type K_get##Name(ok_kernel* k) { return k-> name; }


typedef struct {
    void* data;
    int tag[5];
} ok_tagdata;


double DEGRANGE(double angle);
double RADRANGE(double angle);


void ok_sort_small_matrix(gsl_matrix* matrix, const int column);
void ok_sort_matrix(gsl_matrix* matrix, const int column);
void ok_rsort_matrix(gsl_matrix* matrix, const int column);

void ok_fprintf_matrix(gsl_matrix* matrix, FILE* file, const char* format);
void ok_fprintf_vector(gsl_vector* v, FILE* file, const char* format);
void ok_fprintf_matrix_int(gsl_matrix_int* matrix, FILE* file, const char* format);
void ok_fprintf_vector_int(gsl_vector_int* v, FILE* file, const char* format);
void ok_fprintf_buf(double** buf, FILE* file, const char* format, int rows, int columns);

void ok_fscanf_matrix(gsl_matrix* matrix, FILE* file);
void ok_fscanf_matrix_int(gsl_matrix_int* matrix, FILE* file);
void ok_fscanf_vector(gsl_vector* matrix, FILE* file);
void ok_fscanf_vector_int(gsl_vector_int* matrix, FILE* file);


bool ok_save_matrix(gsl_matrix* matrix, FILE* fid, const char* format);
bool ok_save_matrix_bin(gsl_matrix* matrix, FILE* fid);
bool ok_save_buf(double** matrix, FILE* fid, const char* format, int rows, int cols);
bool ok_save_buf_bin(double** matrix, FILE* fid, int rows, int cols);

gsl_matrix* ok_read_table(FILE* fid);

gsl_vector* ok_vector_resize(gsl_vector* v, int len);
gsl_matrix* ok_matrix_resize(gsl_matrix* v, int rows, int cols);
gsl_vector* ok_vector_resize_pad(gsl_vector* v, int len, double pad);
gsl_matrix* ok_matrix_resize_pad(gsl_matrix* v, int rows, int cols, double pad);

gsl_vector_int* ok_vector_int_resize(gsl_vector_int* v, int len);
gsl_matrix_int* ok_matrix_int_resize(gsl_matrix_int* v, int rows, int cols);

gsl_matrix* ok_matrix_copy(const gsl_matrix* src);
gsl_matrix_int* ok_matrix_int_copy(const gsl_matrix_int* src);
gsl_vector* ok_vector_copy(const gsl_vector* src);
gsl_vector_int* ok_vector_int_copy(const gsl_vector_int* src);
gsl_matrix* ok_buf_to_matrix(double** buf, int rows, int cols);

gsl_matrix* ok_matrix_remove_row(gsl_matrix* m, int row);
gsl_matrix* ok_matrix_remove_column(gsl_matrix* m, int col);
gsl_matrix_int* ok_matrix_int_remove_row(gsl_matrix_int* m, int row);
gsl_matrix_int* ok_matrix_int_remove_column(gsl_matrix_int* m, int col);
gsl_vector* ok_vector_remove(gsl_vector* m, int idx);
gsl_vector_int* ok_vector_int_remove(gsl_vector_int* m, int idx);

void ok_bootstrap_matrix(const gsl_matrix* matrix, gsl_matrix* out, const int sortcol, gsl_rng* r);
void ok_bootstrap_matrix_mean(const gsl_matrix* matrix, int timecol, int valcol, gsl_matrix* out, gsl_rng* r);

gsl_matrix* ok_matrix_filter(gsl_matrix* matrix, const int column, const double filter);
gsl_matrix* ok_matrix_buf_filter(double** matrix, const int rows, const int columns, const int column, const double filter);

int ok_bsearch(double* v, double val, int len);

char* ok_str_copy(const char* src);
char* ok_str_cat(const char* a1, const char* a2);

void ok_avevar(const double* v, int len, double* ave, double* var);

#endif